"""
Fit using the Model interface
=============================

This notebook shows a simple example of using the ``lmfit.Model`` class. For
more information please refer to:
https://lmfit.github.io/lmfit-py/model.html#the-model-class.

"""
import numpy as np
from pandas import Series

from lmfit import Model, Parameter, report_fit

###############################################################################
# The ``Model`` class is a flexible, concise curve fitter. I will illustrate
# fitting example data to an exponential decay.


def decay(t, N, tau):
    return N*np.exp(-t/tau)


###############################################################################
# The parameters are in no particular order. We'll need some example data. I
# will use ``N=7`` and ``tau=3``, and add a little noise.
t = np.linspace(0, 5, num=1000)
np.random.seed(2021)
data = decay(t, 7, 3) + np.random.randn(t.size)

###############################################################################
# **Simplest Usage**
model = Model(decay, independent_vars=['t'])
result = model.fit(data, t=t, N=10, tau=1)

###############################################################################
# The Model infers the parameter names by inspecting the arguments of the
# function, ``decay``. Then I passed the independent variable, ``t``, and
# initial guesses for each parameter. A residual function is automatically
# defined, and a least-squared regression is performed.
#
# We can immediately see the best-fit values:
print(result.values)

###############################################################################
# and use these best-fit parameters for plotting with the ``plot`` function:
result.plot()

###############################################################################
# We can review the best-fit `Parameters` by accessing `result.params`:
result.params.pretty_print()

###############################################################################
# More information about the fit is stored in the result, which is an
# ``lmfit.MimimizerResult`` object (see:
# https://lmfit.github.io/lmfit-py/fitting.html#lmfit.minimizer.MinimizerResult)

###############################################################################
# **Specifying Bounds and Holding Parameters Constant**
#
# Above, the ``Model`` class implicitly builds ``Parameter`` objects from
# keyword arguments of ``fit`` that match the arguments of ``decay``. You can
# build the ``Parameter`` objects explicitly; the following is equivalent.
result = model.fit(data, t=t,
                   N=Parameter('N', value=10),
                   tau=Parameter('tau', value=1))
report_fit(result.params)

###############################################################################
# By building ``Parameter`` objects explicitly, you can specify bounds
# (``min``, ``max``) and set parameters constant (``vary=False``).
result = model.fit(data, t=t,
                   N=Parameter('N', value=7, vary=False),
                   tau=Parameter('tau', value=1, min=0))
report_fit(result.params)

###############################################################################
# **Defining Parameters in Advance**
#
# Passing parameters to ``fit`` can become unwieldy. As an alternative, you
# can extract the parameters from ``model`` like so, set them individually,
# and pass them to ``fit``.
params = model.make_params(N=10, tau={'value': 1, 'min': 0})

result = model.fit(data, params, t=t)
report_fit(result.params)

##############################################################################
# Keyword arguments override ``params``, resetting ``value`` and all other
# properties (``min``, ``max``, ``vary``).
result = model.fit(data, params, t=t, tau=1)
report_fit(result.params)

###############################################################################
# The input parameters are not modified by ``fit``. They can be reused,
# retaining the same initial value. If you want to use the result of one fit
# as the initial guess for the next, simply pass ``params=result.params``.

###############################################################################
# #TODO/FIXME: not sure if there ever way a "helpful exception", but currently
# #it raises a ``ValueError: The input contains nan values``.
#
# #*A Helpful Exception*
#
# #All this implicit magic makes it very easy for the user to neglect to set a
# #parameter. The ``fit`` function checks for this and raises a helpful exception.

# #result = model.fit(data, t=t, tau=1)  # N unspecified

###############################################################################
# #An *extra* parameter that cannot be matched to the model function will
# #throw a ``UserWarning``, but it will not raise, leaving open the possibility
# #of unforeseen extensions calling for some parameters.

###############################################################################
# *Weighted Fits*
#
# Use the ``sigma`` argument to perform a weighted fit. If you prefer to think
# of the fit in term of ``weights``, ``sigma=1/weights``.
weights = np.arange(len(data))
result = model.fit(data, params, t=t, weights=weights)
report_fit(result.params)

###############################################################################
# *Handling Missing Data*
#
# By default, attempting to fit data that includes a ``NaN``, which
# conventionally indicates a "missing" observation, raises a lengthy exception.
# You can choose to ``omit`` (i.e., skip over) missing values instead.
data_with_holes = data.copy()
data_with_holes[[5, 500, 700]] = np.nan  # Replace arbitrary values with NaN.

model = Model(decay, independent_vars=['t'], nan_policy='omit')
result = model.fit(data_with_holes, params, t=t)
report_fit(result.params)

###############################################################################
# If you don't want to ignore missing values, you can set the model to raise
# proactively, checking for missing values before attempting the fit.
#
# Uncomment to see the error
# #model = Model(decay, independent_vars=['t'], nan_policy='raise')
# #result = model.fit(data_with_holes, params, t=t)
#
# The default setting is ``nan_policy='raise'``, which does check for NaNs and
# raises an exception when present.
#
# Null-checking  relies on ``pandas.isnull`` if it is available. If pandas
# cannot be imported, it silently falls back on ``numpy.isnan``.

###############################################################################
# *Data Alignment*
#
# Imagine a collection of time series data with different lengths. It would be
# convenient to define one sufficiently long array ``t`` and use it for each
# time series, regardless of length. ``pandas``
# (https://pandas.pydata.org/pandas-docs/stable/) provides tools for aligning
# indexed data. And, unlike most wrappers to ``scipy.leastsq``, ``Model`` can
# handle pandas objects out of the box, using its data alignment features.
#
# Here I take just a slice of the ``data`` and fit it to the full ``t``. It is
# automatically aligned to the correct section of ``t`` using Series' index.
model = Model(decay, independent_vars=['t'])
truncated_data = Series(data)[200:800]  # data points 200-800
t = Series(t)  # all 1000 points
result = model.fit(truncated_data, params, t=t)
report_fit(result.params)

###############################################################################
# Data with missing entries and an unequal length still aligns properly.
model = Model(decay, independent_vars=['t'], nan_policy='omit')
truncated_data_with_holes = Series(data_with_holes)[200:800]
result = model.fit(truncated_data_with_holes, params, t=t)
report_fit(result.params)



# <examples/doc_model_loadmodelresult2.py>
import os
import sys

import matplotlib.pyplot as plt
import numpy as np

from lmfit.model import load_modelresult

if not os.path.exists('nistgauss_modelresult.sav'):
    os.system(f"{sys.executable} doc_model_savemodelresult2.py")

dat = np.loadtxt('NIST_Gauss2.dat')
x = dat[:, 1]
y = dat[:, 0]

result = load_modelresult('nistgauss_modelresult.sav')
print(result.fit_report())

plt.plot(x, y, 'o')
plt.plot(x, result.best_fit, '-')
plt.show()
# <end examples/doc_model_loadmodelresult2.py>



# <examples/doc_model_savemodel.py>
import numpy as np

from lmfit.model import Model, save_model


def mysine(x, amp, freq, shift):
    return amp * np.sin(x*freq + shift)


sinemodel = Model(mysine)
pars = sinemodel.make_params(amp=1, freq=0.25, shift=0)

save_model(sinemodel, 'sinemodel.sav')
# <end examples/doc_model_savemodel.py>



# <examples/doc_model_loadmodel.py>
import os
import sys

import matplotlib.pyplot as plt
import numpy as np

from lmfit.model import load_model

if not os.path.exists('sinemodel.sav'):
    os.system(f"{sys.executable} doc_model_savemodel.py")


def mysine(x, amp, freq, shift):
    return amp * np.sin(x*freq + shift)


data = np.loadtxt('sinedata.dat')
x = data[:, 0]
y = data[:, 1]

model = load_model('sinemodel.sav', funcdefs={'mysine': mysine})
params = model.make_params(amp=dict(value=3, min=0),
                           freq=0.52,
                           shift=dict(value=0, min=-1, max=1))

result = model.fit(y, params, x=x)
print(result.fit_report())

plt.plot(x, y, 'o')
plt.plot(x, result.best_fit, '-')
plt.show()
# <end examples/doc_model_loadmodel.py>



"""JSON utilities."""

from base64 import b64decode, b64encode
from io import StringIO
import sys
import warnings

import dill
import numpy as np
import uncertainties

HAS_DILL = True

try:
    from pandas import DataFrame, Series, read_json
except ImportError:
    DataFrame = Series = type(NotImplemented)
    read_json = None


pyvers = f'{sys.version_info.major}.{sys.version_info.minor}'


def find_importer(obj):
    """Find importer of an object."""
    oname = obj.__name__
    for modname, module in sys.modules.items():
        if modname.startswith('__main__'):
            continue
        t = getattr(module, oname, None)
        if t is obj:
            return modname
    return None


def import_from(modulepath, objectname):
    """Wrapper for __import__ for nested objects."""
    path = modulepath.split('.')
    top = path.pop(0)
    parent = __import__(top)
    while len(path) > 0:
        parent = getattr(parent, path.pop(0))
    return getattr(parent, objectname)


def encode4js(obj):
    """Prepare an object for JSON encoding.

    It has special handling for many Python types, including:
    - pandas DataFrames and Series
    - NumPy ndarrays
    - complex numbers

    """
    if isinstance(obj, DataFrame):
        return dict(__class__='PDataFrame', value=obj.to_json())
    if isinstance(obj, Series):
        return dict(__class__='PSeries', value=obj.to_json())
    if isinstance(obj, uncertainties.core.AffineScalarFunc):
        return dict(__class__='UFloat', val=obj.nominal_value, err=obj.std_dev)
    if isinstance(obj, np.ndarray):
        if 'complex' in obj.dtype.name:
            val = [(obj.real).tolist(), (obj.imag).tolist()]
        elif obj.dtype.name == 'object':
            val = [encode4js(item) for item in obj]
        else:
            val = obj.flatten().tolist()
        return dict(__class__='NDArray', __shape__=obj.shape,
                    __dtype__=obj.dtype.name, value=val)
    if isinstance(obj, float):
        return float(obj)
    if isinstance(obj, int):
        return int(obj)
    if isinstance(obj, str):
        try:
            return str(obj)
        except UnicodeError:
            return obj
    if isinstance(obj, complex):
        return dict(__class__='Complex', value=(obj.real, obj.imag))
    if isinstance(obj, (tuple, list)):
        ctype = 'List'
        if isinstance(obj, tuple):
            ctype = 'Tuple'
        val = [encode4js(item) for item in obj]
        return dict(__class__=ctype, value=val)
    if isinstance(obj, dict):
        out = dict(__class__='Dict')
        for key, val in obj.items():
            out[encode4js(key)] = encode4js(val)
        return out
    if callable(obj):
        value = str(b64encode(dill.dumps(obj)), 'utf-8')
        return dict(__class__='Callable', __name__=obj.__name__,
                    pyversion=pyvers, value=value,
                    importer=find_importer(obj))
    return obj


def decode4js(obj):
    """Return decoded Python object from encoded object."""
    if not isinstance(obj, dict):
        return obj
    out = obj
    classname = obj.pop('__class__', None)
    if classname is None and isinstance(obj, dict):
        classname = 'dict'
    if classname is None:
        return obj
    if classname == 'Complex':
        out = obj['value'][0] + 1j*obj['value'][1]
    elif classname in ('List', 'Tuple'):
        out = []
        for item in obj['value']:
            out.append(decode4js(item))
        if classname == 'Tuple':
            out = tuple(out)
    elif classname == 'NDArray':
        if obj['__dtype__'].startswith('complex'):
            re = np.fromiter(obj['value'][0], dtype='double')
            im = np.fromiter(obj['value'][1], dtype='double')
            out = re + 1j*im
        elif obj['__dtype__'].startswith('object'):
            val = [decode4js(v) for v in obj['value']]
            out = np.array(val, dtype=obj['__dtype__'])
        else:
            out = np.fromiter(obj['value'], dtype=obj['__dtype__'])
        out.shape = obj['__shape__']
    elif classname == 'PDataFrame' and read_json is not None:
        out = read_json(StringIO(obj['value']))
    elif classname == 'PSeries' and read_json is not None:
        out = read_json(StringIO(obj['value']), typ='series')
    elif classname == 'UFloat':
        out = uncertainties.ufloat(obj['val'], obj['err'])
    elif classname == 'Callable':
        out = obj['__name__']
        try:
            out = import_from(obj['importer'], out)
            unpacked = True
        except (ImportError, AttributeError):
            unpacked = False
        if not unpacked:
            spyvers = obj.get('pyversion', '?')
            if not pyvers == spyvers:
                msg = f"Could not unpack dill-encoded callable '{out}', saved with Python version {spyvers}"
                warnings.warn(msg)

            try:
                out = dill.loads(b64decode(obj['value']))
            except RuntimeError:
                msg = f"Could not unpack dill-encoded callable '{out}`, saved with Python version {spyvers}"
                warnings.warn(msg)

    elif classname in ('Dict', 'dict'):
        out = {}
        for key, val in obj.items():
            out[key] = decode4js(val)
    return out



import numpy as np
from numpy.testing import assert_allclose

from lmfit import Parameters, minimize


def test_bounded_jacobian():
    pars = Parameters()
    pars.add('x0', value=2.0)
    pars.add('x1', value=2.0, min=1.5)

    global jac_count

    jac_count = 0

    def resid(params):
        x0 = params['x0']
        x1 = params['x1']
        return np.array([10 * (x1 - x0*x0), 1-x0])

    def jac(params):
        global jac_count
        jac_count += 1
        x0 = params['x0']
        return np.array([[-20*x0, 10], [-1, 0]])

    out0 = minimize(resid, pars, Dfun=None)

    assert_allclose(out0.params['x0'], 1.2243, rtol=1.0e-4)
    assert_allclose(out0.params['x1'], 1.5000, rtol=1.0e-4)
    assert jac_count == 0

    out1 = minimize(resid, pars, Dfun=jac)

    assert_allclose(out1.params['x0'], 1.2243, rtol=1.0e-4)
    assert_allclose(out1.params['x1'], 1.5000, rtol=1.0e-4)
    assert jac_count > 5


def test_bounded_jacobian_CG():
    pars = Parameters()
    pars.add('x0', value=2.0)
    pars.add('x1', value=2.0, min=1.5)

    global jac_count

    jac_count = 0

    def resid(params):
        x0 = params['x0']
        x1 = params['x1']
        return np.array([10 * (x1 - x0*x0), 1-x0])

    def jac(params):
        global jac_count
        jac_count += 1
        x0 = params['x0']
        # Jacobian of the *error*, i.e. the summed squared residuals
        return 2 * np.sum(np.array([[-20 * x0, 10],
                                    [-1, 0]]).T * resid(params), axis=1)

    out0 = minimize(resid, pars, method='CG')
    assert_allclose(out0.params['x0'], 1.2243, rtol=1.0e-4)
    assert_allclose(out0.params['x1'], 1.5000, rtol=1.0e-4)
    assert jac_count == 0

    out1 = minimize(resid, pars, method='CG', Dfun=jac)
    assert_allclose(out1.params['x0'], 1.2243, rtol=1.0e-4)
    assert_allclose(out1.params['x1'], 1.5000, rtol=1.0e-4)
    assert jac_count > 5



"""Tests for lineshape functions."""

import inspect

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

import lmfit
from lmfit.lineshapes import not_zero, tiny


@pytest.mark.parametrize(
    "value, expected_result",
    [(1, 1.0), (-1, -1.0), (0, tiny), (-0, -tiny)])
def test_not_zero(value, expected_result):
    """Test that not_zero gives the expected results"""
    assert_almost_equal(not_zero(value), expected_result)


@pytest.mark.parametrize("lineshape", lmfit.lineshapes.functions)
def test_no_ZeroDivisionError_and_finite_output(lineshape):
    """Tests for finite output and ZeroDivisionError is not raised."""
    xvals = np.linspace(0, 10, 100)

    func = getattr(lmfit.lineshapes, lineshape)
    assert callable(func)
    sig = inspect.signature(func)

    # set the following function arguments:
    #   x = xvals
    #   center = 0.5*(max(xvals)-min(xvals))
    #   center1 = 0.25*(max(xvals)-min(xvals))
    #   center2 = 0.75*(max(xvals)-min(xvals))
    #   form = default value (i.e., 'linear' or 'bose')
    xvals_mid_range = xvals.mean()
    zero_pars = [par_name for par_name in sig.parameters.keys() if par_name
                 not in ('x', 'form')]

    for par_zero in zero_pars:
        fnc_args = []
        for par in sig.parameters.keys():
            if par == 'x':
                fnc_args.append(xvals)
            elif par == 'center':
                fnc_args.append(0.5*xvals_mid_range)
            elif par == 'center1':
                fnc_args.append(0.25*xvals_mid_range)
            elif par == 'center2':
                fnc_args.append(0.75*xvals_mid_range)
            elif par == par_zero:
                fnc_args.append(0.0)
            else:
                fnc_args.append(sig.parameters[par].default)

        fnc_output = func(*fnc_args)
        assert len(xvals) == len(fnc_output)
        assert np.all(np.isfinite(fnc_output))


@pytest.mark.parametrize("lineshape", lmfit.lineshapes.functions)
def test_x_float_value(lineshape):
    """Test lineshapes when x is not an array but a float."""
    xval = 7.0

    func = getattr(lmfit.lineshapes, lineshape)
    sig = inspect.signature(func)

    fnc_args = [xval]

    for par in [par_name for par_name in sig.parameters.keys()
                if par_name != 'x']:
        fnc_args.append(sig.parameters[par].default)

    fnc_output = func(*fnc_args)
    assert isinstance(fnc_output, float)


rising_form = ['erf', 'logistic', 'atan', 'arctan', 'linear', 'unknown']


@pytest.mark.parametrize("form", rising_form)
@pytest.mark.parametrize("lineshape", ['step', 'rectangle'])
def test_form_argument_step_rectangle(form, lineshape):
    """Test 'form' argument for step- and rectangle-functions."""
    xvals = np.linspace(0, 10, 100)

    func = getattr(lmfit.lineshapes, lineshape)
    sig = inspect.signature(func)

    fnc_args = [xvals]
    for par in [par_name for par_name in sig.parameters.keys()
                if par_name != 'x']:
        if par == 'form':
            fnc_args.append(form)
        else:
            fnc_args.append(sig.parameters[par].default)

    if form == 'unknown':
        msg = r"Invalid value .* for argument .*; should be one of .*"
        with pytest.raises(ValueError, match=msg):
            func(*fnc_args)
    else:
        fnc_output = func(*fnc_args)
        assert len(fnc_output) == len(xvals)


@pytest.mark.parametrize('form', rising_form)
@pytest.mark.parametrize('lineshape', ['step', 'rectangle'])
def test_value_step_rectangle(form, lineshape):
    """Test values at mu1/mu2 for step- and rectangle-functions."""
    func = getattr(lmfit.lineshapes, lineshape)
    # at position mu1 we should be at A/2
    assert_almost_equal(func(0), 0.5)
    # for a rectangular shape we have the same at mu2
    if lineshape == 'rectangle':
        assert_almost_equal(func(1), 0.5)


thermal_form = ['bose', 'maxwell', 'fermi', 'Bose-Einstein', 'unknown']


@pytest.mark.parametrize("form", thermal_form)
def test_form_argument_thermal_distribution(form):
    """Test 'form' argument for thermal_distribution function."""
    xvals = np.linspace(0, 10, 100)

    func = lmfit.lineshapes.thermal_distribution
    sig = inspect.signature(lmfit.lineshapes.thermal_distribution)

    fnc_args = [xvals]
    for par in [par_name for par_name in sig.parameters.keys()
                if par_name != 'x']:
        if par == 'form':
            fnc_args.append(form)
        else:
            fnc_args.append(sig.parameters[par].default)

    if form == 'unknown':
        msg = r"Invalid value .* for argument .*; should be one of .*"
        with pytest.raises(ValueError, match=msg):
            func(*fnc_args)
    else:
        fnc_output = func(*fnc_args)
        assert len(fnc_output) == len(xvals)



"""Tests for the Iteration Callback Function."""

import numpy as np
import pytest

from lmfit.lineshapes import gaussian
from lmfit.minimizer import Minimizer
from lmfit.models import GaussianModel, LinearModel

try:
    import numdifftools  # noqa: F401
    calc_covar_options = [False, True]
except ImportError:
    calc_covar_options = [False]


np.random.seed(7)
x = np.linspace(0, 20, 401)
y = gaussian(x, amplitude=24.56, center=7.6543, sigma=1.23)
y -= 0.20*x + 3.333 + np.random.normal(scale=0.23, size=len(x))
mod = GaussianModel(prefix='peak_') + LinearModel(prefix='bkg_')


def residual(pars, x, data):
    parvals = pars.valuesdict()
    gauss = gaussian(x, parvals['peak_amplitude'], parvals['peak_center'],
                     parvals['peak_sigma'])
    linear = parvals['bkg_slope']*x + parvals['bkg_intercept']
    return data - gauss - linear


pars = mod.make_params(peak_amplitude=21.0, peak_center=7.0,
                       peak_sigma=2.0, bkg_intercept=2, bkg_slope=0.0)

# set bounds for use with 'differential_evolution' and 'brute'
pars['bkg_intercept'].set(min=0, max=10)
pars['bkg_slope'].set(min=-5, max=5)
pars['peak_amplitude'].set(min=20, max=25)
pars['peak_center'].set(min=5, max=10)
pars['peak_sigma'].set(min=0.5, max=2)


def per_iteration(pars, iteration, resid, *args, **kws):
    """Iteration callback, will abort at iteration 23."""
    return iteration == 17


fitmethods = ['ampgo', 'brute', 'basinhopping', 'differential_evolution',
              'leastsq', 'least_squares', 'nelder', 'shgo', 'dual_annealing']


@pytest.mark.parametrize("calc_covar", calc_covar_options)
@pytest.mark.parametrize("method", fitmethods)
def test_itercb_model_class(method, calc_covar):
    """Test the iteration callback for all solvers."""
    out = mod.fit(y, pars, x=x, method=method, iter_cb=per_iteration,
                  calc_covar=calc_covar)

    assert out.nfev == 17
    assert out.aborted
    assert not out.errorbars
    assert not out.success


@pytest.mark.parametrize("calc_covar", calc_covar_options)
@pytest.mark.parametrize("method", fitmethods)
def test_itercb_minimizer_class(method, calc_covar):
    """Test the iteration callback for all solvers."""
    mini = Minimizer(residual, pars, fcn_args=(x, y), iter_cb=per_iteration,
                     calc_covar=calc_covar)
    out = mini.minimize(method=method)
    assert out.nfev == 17
    assert out.aborted
    assert not out.errorbars
    assert not out.success
    if method not in ('nelder', 'differential_evolution'):
        assert mini._abort


fitmethods = ['leastsq', 'least_squares']


@pytest.mark.parametrize("method", fitmethods)
def test_itercb_reset_abort(method):
    """Regression test for GH Issue #756.

    Make sure that ``self._abort`` is reset to ``False`` at the start of each
    fit.

    """
    if method in ('nelder', 'differential_evolution'):
        pytest.xfail("scalar_minimizers behave differently, but shouldn't!!")

    must_stop = True

    def callback(*args, **kwargs):
        return must_stop

    # perform minimization with ``iter_cb``
    out_model = mod.fit(y, pars, x=x, method=method, iter_cb=callback)

    mini = Minimizer(residual, pars, fcn_args=(x, y), iter_cb=callback)
    out_minimizer = mini.minimize(method=method)

    assert out_model.aborted is must_stop
    assert out_model.errorbars is not must_stop
    assert out_model.success is not must_stop
    assert out_minimizer.aborted is must_stop
    assert out_minimizer.errorbars is not must_stop
    assert out_minimizer.success is not must_stop
    assert mini._abort is must_stop

    # perform another minimization now without ``iter_cb``
    must_stop = False
    out_minimizer_no_callback = mini.minimize(method=method)
    assert out_minimizer_no_callback.aborted is must_stop
    assert out_minimizer_no_callback.errorbars is not must_stop
    assert out_minimizer_no_callback.success is not must_stop
    assert mini._abort is must_stop

    # reset to mini._abort to False and call the optimization method directly
    func = getattr(mini, method)
    out_no_callback = func()
    assert out_no_callback.aborted is must_stop
    assert out_no_callback.errorbars is not must_stop
    assert out_no_callback.success is not must_stop
    assert mini._abort is must_stop



import numpy as np

from lmfit.models import ConstantModel, StepModel


def get_data():
    np.random.seed(2021)
    x = np.linspace(0, 10, 201)
    dat = np.ones_like(x)
    dat[:48] = 0.0
    dat[48:77] = np.arange(77-48)/(77.0-48)
    dat = dat + 5e-2*np.random.randn(len(x))
    dat = 110.2 * dat + 12.0
    return x, dat


def test_stepmodel_linear():
    x, y = get_data()
    stepmod = StepModel(form='linear')
    const = ConstantModel()
    pars = stepmod.guess(y, x)
    pars = pars + const.make_params(c=3*y.min())
    mod = stepmod + const

    out = mod.fit(y, pars, x=x)

    assert out.nfev > 5
    assert out.nvarys == 4
    assert out.chisqr > 1
    assert out.params['c'].value > 3
    assert out.params['center'].value > 1
    assert out.params['center'].value < 4
    assert out.params['sigma'].value > 0.5
    assert out.params['sigma'].value < 3.5
    assert out.params['amplitude'].value > 50


def test_stepmodel_erf():
    x, y = get_data()
    stepmod = StepModel(form='linear')
    const = ConstantModel()
    pars = stepmod.guess(y, x)
    pars = pars + const.make_params(c=3*y.min())
    mod = stepmod + const

    out = mod.fit(y, pars, x=x)

    assert out.nfev > 5
    assert out.nvarys == 4
    assert out.chisqr > 1
    assert out.params['c'].value > 3
    assert out.params['center'].value > 1
    assert out.params['center'].value < 4
    assert out.params['amplitude'].value > 50
    assert out.params['sigma'].value > 0.2
    assert out.params['sigma'].value < 1.5



"""
Fit Specifying Different Reduce Function
========================================

The ``reduce_fcn`` specifies how to convert a residual array to a scalar value
for the scalar minimizers. The default value is None (i.e., "sum of squares of
residual") - alternatives are: ``negentropy``, ``neglogcauchy``, or a
user-specified ``callable``. For more information please refer to:
https://lmfit.github.io/lmfit-py/fitting.html#using-the-minimizer-class

Here, we use as an example the Student's t log-likelihood for robust fitting
of data with outliers.

"""
import matplotlib.pyplot as plt
import numpy as np

import lmfit


def resid(params, x, ydata):
    decay = params['decay'].value
    offset = params['offset'].value
    omega = params['omega'].value
    amp = params['amp'].value

    y_model = offset + amp * np.sin(x*omega) * np.exp(-x/decay)
    return y_model - ydata


###############################################################################
# Generate synthetic data with noise/outliers and initialize fitting Parameters:
decay = 5
offset = 1.0
amp = 2.0
omega = 4.0

np.random.seed(2)
x = np.linspace(0, 10, 101)
y = offset + amp * np.sin(omega*x) * np.exp(-x/decay)
yn = y + np.random.normal(size=y.size, scale=0.250)

outliers = np.random.randint(int(len(x)/3.0), len(x), int(len(x)/12))
yn[outliers] += 5*np.random.random(len(outliers))

params = lmfit.create_params(offset=2.0, omega=3.3, amp=2.5,
                             decay=dict(value=1, min=0))

###############################################################################
# Perform fits using the ``L-BFGS-B`` method with different ``reduce_fcn``:
method = 'L-BFGS-B'
o1 = lmfit.minimize(resid, params, args=(x, yn), method=method)
print("# Fit using sum of squares:\n")
lmfit.report_fit(o1)

###############################################################################
o2 = lmfit.minimize(resid, params, args=(x, yn), method=method,
                    reduce_fcn='neglogcauchy')
print("\n\n# Robust Fit, using log-likelihood with Cauchy PDF:\n")
lmfit.report_fit(o2)

###############################################################################
plt.plot(x, y, 'o', label='true function')
plt.plot(x, yn, '--*', label='with noise+outliers')
plt.plot(x, yn+o1.residual, '-', label='sum of squares fit')
plt.plot(x, yn+o2.residual, '-', label='robust fit')
plt.legend()
plt.show()



# <examples/doc_builtinmodels_peakmodels.py>
import matplotlib.pyplot as plt
from numpy import loadtxt

from lmfit.models import GaussianModel, LorentzianModel, VoigtModel

data = loadtxt('test_peak.dat')
x = data[:, 0]
y = data[:, 1]


# Gaussian model
mod = GaussianModel()
pars = mod.guess(y, x=x)
out = mod.fit(y, pars, x=x)

print(out.fit_report(correl_mode='table'))

plt.plot(x, y)
plt.plot(x, out.best_fit, '-', label='Gaussian Model')
plt.legend()
plt.show()


# Lorentzian model
mod = LorentzianModel()
pars = mod.guess(y, x=x)
out = mod.fit(y, pars, x=x)

print(out.fit_report(correl_mode='table'))

plt.figure()
plt.plot(x, y, '-')
plt.plot(x, out.best_fit, '-', label='Lorentzian Model')
plt.legend()
plt.show()


# Voigt model
mod = VoigtModel()
pars = mod.guess(y, x=x)
out = mod.fit(y, pars, x=x)

print(out.fit_report(correl_mode='table'))

fig, axes = plt.subplots(1, 2, figsize=(12.8, 4.8))

axes[0].plot(x, y, '-')
axes[0].plot(x, out.best_fit, '-', label='Voigt Model\ngamma constrained')
axes[0].legend()

# allow the gamma parameter to vary in the fit
pars['gamma'].vary = True
out_gamma = mod.fit(y, pars, x=x)
print(out.fit_report(correl_mode='table'))

axes[1].plot(x, y, '-')
axes[1].plot(x, out_gamma.best_fit, '-', label='Voigt Model\ngamma unconstrained')
axes[1].legend()

plt.show()
# <end examples/doc_builtinmodels_peakmodels.py>



import numpy as np

from lmfit.lineshapes import gaussian
from lmfit.models import GaussianModel


def test_default_inputs_gauss():
    area = 1
    cen = 0
    std = 0.2
    x = np.arange(-3, 3, 0.01)
    y = gaussian(x, area, cen, std)

    g = GaussianModel()

    fit_option1 = {'xtol': 1e-2}
    result1 = g.fit(y, x=x, amplitude=1, center=0, sigma=0.5,
                    max_nfev=5000, fit_kws=fit_option1)

    fit_option2 = {'xtol': 1e-6}
    result2 = g.fit(y, x=x, amplitude=1, center=0, sigma=0.5,
                    max_nfev=5000, fit_kws=fit_option2)

    assert result1.values != result2.values



# <examples/doc_fitting_emcee.py>
import numpy as np

import lmfit

try:
    import matplotlib.pyplot as plt
    HASPYLAB = True
except ImportError:
    HASPYLAB = False

try:
    import corner
    HASCORNER = True
except ImportError:
    HASCORNER = False

x = np.linspace(1, 10, 250)
np.random.seed(0)
y = (3.0*np.exp(-x/2) - 5.0*np.exp(-(x-0.1) / 10.) +
     0.1*np.random.randn(x.size))

p = lmfit.Parameters()
p.add_many(('a1', 4), ('a2', 4), ('t1', 3), ('t2', 3., True))


def residual(p):
    v = p.valuesdict()
    return v['a1']*np.exp(-x/v['t1']) + v['a2']*np.exp(-(x-0.1) / v['t2']) - y


mi = lmfit.minimize(residual, p, method='nelder', nan_policy='omit')
lmfit.printfuncs.report_fit(mi.params, min_correl=0.5)
if HASPYLAB:
    plt.figure()
    plt.plot(x, y, 'o')
    plt.plot(x, residual(mi.params) + y, label='best fit')
    plt.legend()
    plt.show()

# Place bounds on the ln(sigma) parameter that emcee will automatically add
# to estimate the true uncertainty in the data since is_weighted=False
mi.params.add('__lnsigma', value=np.log(0.1), min=np.log(0.001), max=np.log(2))

res = lmfit.minimize(residual, method='emcee', nan_policy='omit', burn=300,
                     steps=1000, thin=20, params=mi.params, is_weighted=False,
                     progress=False)

if HASPYLAB and HASCORNER:
    emcee_corner = corner.corner(res.flatchain, labels=res.var_names,
                                 truths=list(res.params.valuesdict().values()))
    plt.show()

if HASPYLAB:
    plt.plot(res.acceptance_fraction, 'o')
    plt.xlabel('walker')
    plt.ylabel('acceptance fraction')
    plt.show()

if hasattr(res, "acor"):
    print("Autocorrelation time for the parameters:")
    print("----------------------------------------")
    for i, par in enumerate(p):
        print(par, res.acor[i])

print("\nmedian of posterior probability distribution")
print('--------------------------------------------')
lmfit.report_fit(res.params)


# find the maximum likelihood solution
highest_prob = np.argmax(res.lnprob)
hp_loc = np.unravel_index(highest_prob, res.lnprob.shape)
mle_soln = res.chain[hp_loc]
for i, par in enumerate(p):
    p[par].value = mle_soln[i]

print('\nMaximum Likelihood Estimation from emcee       ')
print('-------------------------------------------------')
print('Parameter  MLE Value   Median Value   Uncertainty')
fmt = '  {:5s}  {:11.5f} {:11.5f}   {:11.5f}'.format
for name, param in p.items():
    print(fmt(name, param.value, res.params[name].value,
              res.params[name].stderr))

if HASPYLAB:
    plt.figure()
    plt.plot(x, y, 'o')
    plt.plot(x, residual(mi.params) + y, label='Nelder-Mead')
    plt.plot(x, residual(res.params) + y, '--', label='emcee')
    plt.legend()
    plt.show()

print('\nError Estimates from emcee    ')
print('------------------------------------------------------')
print('Parameter  -2sigma  -1sigma   median  +1sigma  +2sigma ')

for name in p.keys():
    quantiles = np.percentile(res.flatchain[name],
                              [2.275, 15.865, 50, 84.135, 97.275])
    median = quantiles[2]
    err_m2 = quantiles[0] - median
    err_m1 = quantiles[1] - median
    err_p1 = quantiles[3] - median
    err_p2 = quantiles[4] - median
    fmt = '  {:5s}   {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f}'.format
    print(fmt(name, err_m2, err_m1, median, err_p1, err_p2))



"""Tests for the Parameters class."""

from copy import copy, deepcopy
import os
import pickle

import numpy as np
from numpy.testing import assert_allclose
import pytest

import lmfit
from lmfit.models import VoigtModel


@pytest.fixture
def parameters():
    """Initialize a Parameters class for tests."""
    pars = lmfit.Parameters()
    pars.add(lmfit.Parameter(name='a', value=10.0, vary=True, min=-100.0,
                             max=100.0, expr=None, brute_step=5.0,
                             user_data=1))
    pars.add(lmfit.Parameter(name='b', value=0.0, vary=True, min=-250.0,
                             max=250.0, expr="2.0*a", brute_step=25.0,
                             user_data={'test': 123}))
    exp_attr_values_A = ('a', 10.0, True, -100.0, 100.0, None, 5.0, 1)
    exp_attr_values_B = ('b', 20.0, False, -250.0, 250.0, "2.0*a", 25.0, {'test': 123})
    assert_parameter_attributes(pars['a'], exp_attr_values_A)
    assert_parameter_attributes(pars['b'], exp_attr_values_B)
    return pars, exp_attr_values_A, exp_attr_values_B


def assert_parameter_attributes(par, expected):
    """Assert that parameter attributes have the expected values."""
    par_attr_values = (par.name, par._val, par.vary, par.min, par.max,
                       par._expr, par.brute_step, par.user_data)
    assert par_attr_values == expected


def test_check_ast_errors():
    """Assert that an exception is raised upon AST errors."""
    pars = lmfit.Parameters()

    msg = r"at expr='<_?ast.Module object at"
    with pytest.raises(NameError, match=msg):
        pars.add('par1', expr='2.0*par2')


def test_parameters_init_with_usersyms():
    """Test for initialization of the Parameters class with usersyms."""
    pars = lmfit.Parameters(usersyms={'test': np.sin})
    assert 'test' in pars._asteval.symtable


def test_parameters_copy(parameters):
    """Tests for copying a Parameters class; all use the __deepcopy__ method."""
    pars, exp_attr_values_A, exp_attr_values_B = parameters

    copy_pars = copy(pars)
    pars_copy = pars.copy()
    pars__copy__ = pars.__copy__()

    # modifying the original parameters should not modify the copies
    pars['a'].set(value=100)
    pars['b'].user_data['test'] = 456

    for copied in [copy_pars, pars_copy, pars__copy__]:
        assert isinstance(copied, lmfit.Parameters)
        assert copied != pars
        assert copied._asteval is not None
        assert copied._asteval.symtable is not None
        assert_parameter_attributes(copied['a'], exp_attr_values_A)
        assert_parameter_attributes(copied['b'], exp_attr_values_B)


def test_parameters_deepcopy(parameters):
    """Tests for deepcopy of a Parameters class."""
    pars, _, _ = parameters

    deepcopy_pars = deepcopy(pars)
    assert isinstance(deepcopy_pars, lmfit.Parameters)
    assert deepcopy_pars == pars

    # check that we can add a symbol to the interpreter
    pars['b'].expr = 'sin(1)'
    pars['b'].value = 10
    assert_allclose(pars['b'].value, np.sin(1))
    assert_allclose(pars._asteval.symtable['b'], np.sin(1))

    # check that the symbols in the interpreter are still the same after
    # deepcopying
    pars, exp_attr_values_A, exp_attr_values_B = parameters
    deepcopy_pars = deepcopy(pars)

    unique_symbols_pars = pars._asteval.user_defined_symbols()
    unique_symbols_copied = deepcopy_pars._asteval.user_defined_symbols()
    assert unique_symbols_copied == unique_symbols_pars

    for unique_symbol in unique_symbols_copied:
        if pars._asteval.symtable[unique_symbol] is not np.nan:
            assert (pars._asteval.symtable[unique_symbol] ==
                    deepcopy_pars._asteval.symtable[unique_symbol])


def test_parameters_deepcopy_subclass():
    """Test that a subclass of parameters is preserved when performing a deepcopy"""
    class ParametersSubclass(lmfit.Parameters):
        pass

    parameters = ParametersSubclass()
    assert isinstance(parameters, ParametersSubclass)

    parameterscopy = deepcopy(parameters)
    assert isinstance(parameterscopy, ParametersSubclass)


def test_parameters_update(parameters):
    """Tests for updating a Parameters class."""
    pars, exp_attr_values_A, exp_attr_values_B = parameters

    msg = r"'test' is not a Parameters object"
    with pytest.raises(ValueError, match=msg):
        pars.update('test')

    pars2 = lmfit.Parameters()
    pars2.add(lmfit.Parameter(name='c', value=7.0, vary=True, min=-70.0,
                              max=70.0, expr=None, brute_step=0.7,
                              user_data=7))
    exp_attr_values_C = ('c', 7.0, True, -70.0, 70.0, None, 0.7, 7)

    pars_updated = pars.update(pars2)

    assert_parameter_attributes(pars_updated['a'], exp_attr_values_A)
    assert_parameter_attributes(pars_updated['b'], exp_attr_values_B)
    assert_parameter_attributes(pars_updated['c'], exp_attr_values_C)


def test_parameters__setitem__(parameters):
    """Tests for __setitem__ method of a Parameters class."""
    pars, _, exp_attr_values_B = parameters

    msg = r"'10' is not a valid Parameters name"
    with pytest.raises(KeyError, match=msg):
        pars.__setitem__('10', None)

    msg = r"'not_a_parameter' is not a Parameter"
    with pytest.raises(ValueError, match=msg):
        pars.__setitem__('a', 'not_a_parameter')

    par = lmfit.Parameter('b', value=10, min=-25.0, brute_step=1)
    pars.__setitem__('b', par)

    exp_attr_values_B = ('b', 10, True, -25.0, np.inf, None, 1, None)
    assert_parameter_attributes(pars['b'], exp_attr_values_B)


def test_parameters__add__(parameters):
    """Test the __add__ magic method."""
    pars, exp_attr_values_A, exp_attr_values_B = parameters

    msg = r"'other' is not a Parameters object"
    with pytest.raises(ValueError, match=msg):
        _ = pars + 'other'

    pars2 = lmfit.Parameters()
    pars2.add_many(('c', 1., True, None, None, None),
                   ('d', 2., True, None, None, None))
    exp_attr_values_C = ('c', 1, True, -np.inf, np.inf, None, None, None)
    exp_attr_values_D = ('d', 2, True, -np.inf, np.inf, None, None, None)

    pars_added = pars + pars2

    assert_parameter_attributes(pars_added['a'], exp_attr_values_A)
    assert_parameter_attributes(pars_added['b'], exp_attr_values_B)
    assert_parameter_attributes(pars_added['c'], exp_attr_values_C)
    assert_parameter_attributes(pars_added['d'], exp_attr_values_D)


def test_parameters__iadd__(parameters):
    """Test the __iadd__ magic method."""
    pars, exp_attr_values_A, exp_attr_values_B = parameters

    msg = r"'other' is not a Parameters object"
    with pytest.raises(ValueError, match=msg):
        pars += 'other'

    pars2 = lmfit.Parameters()
    pars2.add_many(('c', 1., True, None, None, None),
                   ('d', 2., True, None, None, None))
    exp_attr_values_C = ('c', 1, True, -np.inf, np.inf, None, None, None)
    exp_attr_values_D = ('d', 2, True, -np.inf, np.inf, None, None, None)

    pars += pars2

    assert_parameter_attributes(pars['a'], exp_attr_values_A)
    assert_parameter_attributes(pars['b'], exp_attr_values_B)
    assert_parameter_attributes(pars['c'], exp_attr_values_C)
    assert_parameter_attributes(pars['d'], exp_attr_values_D)


def test_parameters_add_with_symtable():
    """Regression test for GitHub Issue 607."""
    pars1 = lmfit.Parameters()
    pars1.add('a', value=1.0)

    def half(x):
        return 0.5*x

    pars2 = lmfit.Parameters(usersyms={"half": half})
    pars2.add("b", value=3.0)
    pars2.add("c", expr="half(b)")

    params = pars1 + pars2
    assert_allclose(params['c'].value, 1.5)

    params = pars2 + pars1
    assert_allclose(params['c'].value, 1.5)

    params = deepcopy(pars1)
    params.update(pars2)
    assert_allclose(params['c'].value, 1.5)

    pars1 += pars2
    assert_allclose(params['c'].value, 1.5)


def test_parameters__array__(parameters):
    """Test the __array__ magic method."""
    pars, _, _ = parameters

    assert_allclose(np.array(pars), np.array([10.0, 20.0]))


def test_parameters__reduce__(parameters):
    """Test the __reduce__ magic method."""
    pars, _, _ = parameters
    reduced = pars.__reduce__()

    assert isinstance(reduced[2], dict)
    assert 'unique_symbols' in reduced[2].keys()
    assert reduced[2]['unique_symbols']['b'] == 20
    assert 'params' in reduced[2].keys()
    assert isinstance(reduced[2]['params'][0], lmfit.Parameter)


def test_parameters__setstate__(parameters):
    """Test the __setstate__ magic method."""
    pars, exp_attr_values_A, exp_attr_values_B = parameters
    reduced = pars.__reduce__()

    pars_setstate = lmfit.Parameters()
    pars_setstate.__setstate__(reduced[2])

    assert isinstance(pars_setstate, lmfit.Parameters)
    assert_parameter_attributes(pars_setstate['a'], exp_attr_values_A)
    assert_parameter_attributes(pars_setstate['b'], exp_attr_values_B)


def test_pickle_parameters():
    """Test that we can pickle a Parameters object."""
    p = lmfit.Parameters()
    p.add('a', 10, True, 0, 100)
    p.add('b', 10, True, 0, 100, 'a * sin(1)')
    p.update_constraints()
    p._asteval.symtable['abc'] = '2 * 3.142'

    pkl = pickle.dumps(p, -1)
    q = pickle.loads(pkl)

    q.update_constraints()
    assert p == q
    assert p is not q

    # now test if the asteval machinery survived
    assert q._asteval.symtable['abc'] == '2 * 3.142'

    # check that unpickling of Parameters is not affected by expr that
    # refer to Parameter that are added later on. In the following
    # example var_0.expr refers to var_1, which is a Parameter later
    # on in the Parameters dictionary.
    p = lmfit.Parameters()
    p.add('var_0', value=1)
    p.add('var_1', value=2)
    p['var_0'].expr = 'var_1'
    pkl = pickle.dumps(p)
    q = pickle.loads(pkl)


def test_parameters_eval(parameters):
    """Test the eval method."""
    pars, _, _ = parameters
    evaluated = pars.eval('10.0*a+b')
    assert_allclose(evaluated, 120)

    # check that eval() works with usersyms and parameter values
    def myfun(x):
        return 2.0 * x

    pars2 = lmfit.Parameters(usersyms={"myfun": myfun})
    pars2.add('a', value=4.0)
    pars2.add('b', value=3.0)
    assert_allclose(pars2.eval('myfun(2.0) * a'), 16)
    assert_allclose(pars2.eval('b / myfun(3.0)'), 0.5)


def test_parameters_pretty_repr(parameters):
    """Test the pretty_repr method."""
    pars, _, _ = parameters
    output = pars.pretty_repr()
    output_oneline = pars.pretty_repr(oneline=True)

    split_output = output.split('\n')
    assert len(split_output) == 5
    assert 'Parameters' in split_output[0]
    assert "Parameter 'a'" in split_output[1]
    assert "Parameter 'b'" in split_output[2]

    oneliner = ("Parameters([('a', <Parameter 'a', value=10.0, "
                "bounds=[-100.0:100.0], brute_step=5.0>), ('b', <Parameter "
                "'b', value=20.0, bounds=[-250.0:250.0], expr='2.0*a', "
                "brute_step=25.0>)])")
    assert output_oneline == oneliner


def test_parameters_pretty_print(parameters, capsys):
    """Test the pretty_print method."""
    pars, _, _ = parameters

    # oneliner
    pars.pretty_print(oneline=True)
    captured = capsys.readouterr()
    oneliner = ("Parameters([('a', <Parameter 'a', value=10.0, "
                "bounds=[-100.0:100.0], brute_step=5.0>), ('b', <Parameter "
                "'b', value=20.0, bounds=[-250.0:250.0], expr='2.0*a', "
                "brute_step=25.0>)])")
    assert oneliner in captured.out

    # default
    pars.pretty_print()
    captured = capsys.readouterr()
    captured_split = captured.out.split('\n')
    assert len(captured_split) == 4
    header = ('Name     Value      Min      Max   Stderr     Vary     '
              'Expr Brute_Step')
    assert captured_split[0] == header

    # specify columnwidth
    pars.pretty_print(colwidth=12)
    captured = capsys.readouterr()
    captured_split = captured.out.split('\n')
    header = ('Name         Value          Min          Max       Stderr     '
              '    Vary         Expr   Brute_Step')
    assert captured_split[0] == header

    # specify columns
    pars['a'].stderr = 0.01
    pars.pretty_print(columns=['value', 'min', 'max', 'stderr'])
    captured = capsys.readouterr()
    captured_split = captured.out.split('\n')
    assert captured_split[0] == 'Name     Value      Min      Max   Stderr'
    assert captured_split[1] == 'a        10     -100      100     0.01'
    assert captured_split[2] == 'b        20     -250      250     None'

    # specify fmt
    pars.pretty_print(fmt='e', columns=['value', 'min', 'max'])
    captured = capsys.readouterr()
    captured_split = captured.out.split('\n')
    assert captured_split[0] == 'Name     Value      Min      Max'
    assert captured_split[1] == 'a  1.0000e+01 -1.0000e+02 1.0000e+02'
    assert captured_split[2] == 'b  2.0000e+01 -2.5000e+02 2.5000e+02'

    # specify precision
    pars.pretty_print(precision=2, fmt='e', columns=['value', 'min', 'max'])
    captured = capsys.readouterr()
    captured_split = captured.out.split('\n')
    assert captured_split[0] == 'Name     Value      Min      Max'
    assert captured_split[1] == 'a  1.00e+01 -1.00e+02 1.00e+02'
    assert captured_split[2] == 'b  2.00e+01 -2.50e+02 2.50e+02'


def test_parameters__repr_html_(parameters):
    """Test _repr_html method to generate HTML table for Parameters class."""
    pars, _, _ = parameters
    repr_html = pars._repr_html_()

    assert isinstance(repr_html, str)
    assert '<table class="jp-toc-ignore"><caption>Parameters</caption>' in repr_html


def test_parameters_add():
    """Tests for adding a Parameter to the Parameters class."""
    pars = lmfit.Parameters()
    pars_from_par = lmfit.Parameters()

    pars.add('a')
    exp_attr_values_A = ('a', -np.inf, True, -np.inf, np.inf, None, None, None)
    assert_parameter_attributes(pars['a'], exp_attr_values_A)

    pars_from_par.add(lmfit.Parameter('a'))
    assert pars_from_par == pars

    pars.add('b', value=1, vary=False, min=-5.0, max=5.0, brute_step=0.1)
    exp_attr_values_B = ('b', 1.0, False, -5.0, 5.0, None, 0.1, None)
    assert_parameter_attributes(pars['b'], exp_attr_values_B)

    pars_from_par.add(lmfit.Parameter('b', value=1, vary=False, min=-5.0,
                                      max=5.0, brute_step=0.1))
    assert pars_from_par == pars


def test_add_params_expr_outoforder():
    """Regression test for GitHub Issue 560."""
    params1 = lmfit.Parameters()
    params1.add("a", value=1.0)

    params2 = lmfit.Parameters()
    params2.add("b", value=1.0)
    params2.add("c", value=2.0)
    params2['b'].expr = 'c/2'

    params = params1 + params2
    assert 'b' in params
    assert_allclose(params['b'].value, 1.0)


def test_parameters_add_many():
    """Tests for add_many method."""
    a = lmfit.Parameter('a', 1)
    b = lmfit.Parameter('b', 2)

    par = lmfit.Parameters()
    par.add_many(a, b)

    par_with_tuples = lmfit.Parameters()
    par_with_tuples.add_many(('a', 1), ('b', 2))

    assert list(par.keys()) == ['a', 'b']
    assert par == par_with_tuples


def test_parameters_valuesdict(parameters):
    """Test for valuesdict method."""
    pars, _, _ = parameters
    vals_dict = pars.valuesdict()

    assert isinstance(vals_dict, dict)
    assert_allclose(vals_dict['a'], pars['a'].value)
    assert_allclose(vals_dict['b'], pars['b'].value)


def test_dumps_loads_parameters(parameters):
    """Test for dumps and loads methods for a Parameters class."""
    pars, _, _ = parameters

    dumps = pars.dumps()
    assert isinstance(dumps, str)
    newpars = lmfit.Parameters().loads(dumps)
    assert newpars == pars

    newpars['a'].value = 100.0
    assert_allclose(newpars['b'].value, 200.0)


def test_dump_load_parameters(parameters):
    """Test for dump and load methods for a Parameters class."""
    pars, _, _ = parameters

    with open('parameters.sav', 'w') as outfile:
        pars.dump(outfile)

    with open('parameters.sav') as infile:
        newpars = pars.load(infile)

    assert newpars == pars
    newpars['a'].value = 100.0
    assert_allclose(newpars['b'].value, 200.0)


def test_dumps_loads_parameters_usersyms():
    """Test for dumps/loads methods for a Parameters class with usersyms."""
    def half(x):
        return 0.5*x

    pars = lmfit.Parameters(usersyms={"half": half, 'my_func': np.sqrt})
    pars.add(lmfit.Parameter(name='a', value=9.0, min=-100.0, max=100.0))
    pars.add(lmfit.Parameter(name='b', value=100.0, min=-250.0, max=250.0))
    pars.add("c", expr="half(b) + my_func(a)")

    dumps = pars.dumps()
    assert isinstance(dumps, str)
    assert '"half": {' in dumps
    assert '"my_func": {' in dumps

    newpars = lmfit.Parameters().loads(dumps)
    assert 'half' in newpars._asteval.symtable
    assert 'my_func' in newpars._asteval.symtable
    assert_allclose(newpars['a'].value, 9.0)
    assert_allclose(newpars['b'].value, 100.0)

    # within the py.test environment the encoding of the function 'half' does
    # not work correctly as it is changed from <function half at 0x?????????>"
    # to "<function test_dumps_loads_parameters_usersyms.<locals>.half at 0x?????????>
    # This result in the "importer" to be set to None and the final "decode4js"
    # does not do the correct thing.
    #
    # Of note, this is only an issue within the py.test framework and it DOES
    # work correctly in a normal Python interpreter. Also, it isn't an issue
    # when DILL is used, so in that case the two asserts below will pass.
    assert newpars == pars
    assert_allclose(newpars['c'].value, 53.0)


def test_parameters_expr_and_constraints():
    """Regression tests for GitHub Issue #265. Test that parameters are re-
    evaluated if they have bounds and expr.

    """
    p = lmfit.Parameters()
    p.add(lmfit.Parameter('a', 10, True))
    p.add(lmfit.Parameter('b', 10, True, 0, 20))

    assert_allclose(p['b'].min, 0)
    assert_allclose(p['b'].max, 20)

    p['a'].expr = '2 * b'
    assert_allclose(p['a'].value, 20)

    p['b'].value = 15
    assert_allclose(p['b'].value, 15)
    assert_allclose(p['a'].value, 30)

    p['b'].value = 30
    assert_allclose(p['b'].value, 20)
    assert_allclose(p['a'].value, 40)


def test_parameters_usersyms():
    """Test for passing usersyms to Parameters()."""
    def myfun(x):
        return x**3

    params = lmfit.Parameters(usersyms={"myfun": myfun})
    params.add("a", value=2.3)
    params.add("b", expr="myfun(a)")

    np.random.seed(2020)
    xx = np.linspace(0, 1, 10)
    yy = 3 * xx + np.random.normal(scale=0.002, size=xx.size)

    model = lmfit.Model(lambda x, a: a * x)
    result = model.fit(yy, params=params, x=xx)
    assert_allclose(result.params['a'].value, 3.0, rtol=1e-3)
    assert (result.nfev > 3 and result.nfev < 300)


def test_parameters_expr_with_bounds():
    """Test Parameters using an expression with bounds, without value."""
    pars = lmfit.Parameters()
    pars.add('c1', value=0.2)
    pars.add('c2', value=0.2)
    pars.add('c3', value=0.2)
    pars.add('csum', value=0.8)

    # this should not raise TypeError:
    pars.add('c4', expr='csum-c1-c2-c3', min=0, max=1)
    assert_allclose(pars['c4'].value, 0.2)


def test_invalid_expr_exceptions():
    """Regression test for GitHub Issue #486: check that an exception is
    raised for invalid expressions.

    """
    p1 = lmfit.Parameters()
    p1.add('t', 2.0, min=0.0, max=5.0)
    p1.add('x', 10.0)

    with pytest.raises(SyntaxError):
        p1.add('y', expr='x*t + sqrt(t)/')
    assert len(p1['y']._expr_eval.error) > 0

    p1.add('y', expr='x*t + sqrt(t)/3.0')
    p1['y'].set(expr='x*3.0 + t**2')
    assert 'x*3' in p1['y'].expr
    assert len(p1['y']._expr_eval.error) == 0

    with pytest.raises(SyntaxError):
        p1['y'].set(expr='t+')
    assert len(p1['y']._expr_eval.error) > 0
    assert_allclose(p1['y'].value, 34.0)


def test_create_params():
    """Tests for create_params() function."""
    pars1 = lmfit.create_params(a=8, b=9,
                                c=dict(value=3, min=0, max=10),
                                d=dict(expr='a+b/c'),
                                e=dict(value=10000, brute_step=4))

    assert pars1['a'].value == 8
    assert pars1['b'].value == 9
    assert pars1['c'].value == 3
    assert pars1['c'].min == 0
    assert pars1['c'].max == 10
    assert pars1['d'].expr == 'a+b/c'
    assert pars1['d'].value == 11
    assert pars1['e'].value == 10000
    assert pars1['e'].brute_step == 4


def test_unset_constrained_param():
    """test 'unsetting' a constrained parameter by
    just setting `param.vary = True`

    """
    data = np.loadtxt(os.path.join(os.path.dirname(__file__), '..',
                                   'examples', 'test_peak.dat'))
    x = data[:, 0]
    y = data[:, 1]

    # initial fit
    mod = VoigtModel()
    params = mod.guess(y, x=x)
    out1 = mod.fit(y, params, x=x)

    assert out1.nvarys == 3
    assert out1.chisqr < 20.0

    # now just gamma to vary
    params['gamma'].vary = True
    out2 = mod.fit(y, params, x=x)

    assert out2.nvarys == 4
    assert out2.chisqr < out1.chisqr
    assert out2.rsquared > out1.rsquared
    assert out2.params['gamma'].correl['sigma'] < -0.6



#!/bin/env python
"""

Wrapper for emcee

"""

import numpy as np
from .minimizer import _nan_policy

# check for emcee
try:
    import emcee
    from emcee.autocorr import AutocorrError
    HAS_EMCEE = int(emcee.__version__[0]) >= 3
except ImportError:
    HAS_EMCEE = False

# check for pandas
try:
    import pandas as pd
    from pandas import isnull
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False
    isnull = np.isnan

def _lnprob(minimizer, theta, userfcn, params, var_names, bounds, userargs=(),
            userkws=None, float_behavior='posterior', is_weighted=True,
            nan_policy='raise'):
    """Calculate the log-posterior probability.

    See the `Minimizer.emcee` method for more details.

    Parameters
    ----------
    minimizer : minimizer
        Minimizer instance
    theta : sequence
        Float parameter values (only those being varied).
    userfcn : callable
        User objective function.
    params : Parameters
        The entire set of Parameters.
    var_names : list
        The names of the parameters that are varying.
    bounds : numpy.ndarray
        Lower and upper bounds of parameters, with shape
        ``(nvarys, 2)``.
    userargs : tuple, optional
        Extra positional arguments required for user objective function.
    userkws : dict, optional
        Extra keyword arguments required for user objective function.
    float_behavior : {'posterior', 'chi2'}, optional
        Specifies meaning of objective when it returns a float. Use
        `'posterior'` if objective function returns a log-posterior
        probability (default) or `'chi2'` if it returns a chi2 value.
    is_weighted : bool, optional
        If `userfcn` returns a vector of residuals then `is_weighted`
        (default is True) specifies if the residuals have been weighted
        by data uncertainties.
    nan_policy : {'raise', 'propagate', 'omit'}, optional
        Specifies action if `userfcn` returns NaN values. Use `'raise'`
        (default) to raise a `ValueError`, `'propagate'` to use values
        as-is, or `'omit'` to filter out the non-finite values.

    Returns
    -------
    lnprob : float
        Log posterior probability.

    """
    # the comparison has to be done on theta and bounds. DO NOT inject theta
    # values into Parameters, then compare Parameters values to the bounds.
    # Parameters values are clipped to stay within bounds.
    if np.any(theta > bounds[:, 1]) or np.any(theta < bounds[:, 0]):
        return -np.inf
    for name, val in zip(var_names, theta):
        params[name].value = val
    userkwargs = {}
    if userkws is not None:
        userkwargs = userkws
    # update the constraints
    params.update_constraints()
    # now calculate the log-likelihood
    out = userfcn(params, *userargs, **userkwargs)
    minimizer.result.nfev += 1
    if callable(minimizer.iter_cb):
        abort = minimizer.iter_cb(params, minimizer.result.nfev, out,
                             *userargs, **userkwargs)
        minimizer._abort = minimizer._abort or abort
    if minimizer._abort:
        minimizer.result.residual = out
        minimizer._lastpos = theta
        raise AbortFitException("fit aborted by user.")
    else:
        out = _nan_policy(np.asarray(out).ravel(),
                          nan_policy=minimizer.nan_policy)
    lnprob = np.asarray(out).ravel()
    if len(lnprob) == 0:
        lnprob = np.array([-1.e100])
    if lnprob.size > 1:
        # objective function returns a vector of residuals
        if '__lnsigma' in params and not is_weighted:
            # marginalise over a constant data uncertainty
            __lnsigma = params['__lnsigma'].value
            c = np.log(2 * np.pi) + 2 * __lnsigma
            lnprob = -0.5 * np.sum((lnprob / np.exp(__lnsigma)) ** 2 + c)
        else:
            lnprob = -0.5 * (lnprob * lnprob).sum()
    else:
        # objective function returns a single value.
        # use float_behaviour to figure out if the value is posterior or chi2
        if float_behavior == 'posterior':
            pass
        elif float_behavior == 'chi2':
            lnprob *= -0.5
        else:
            raise ValueError("float_behaviour must be either 'posterior' "
                             "or 'chi2' " + float_behavior)
    return lnprob

def emcee(minimizer, params=None, steps=1000, nwalkers=100, burn=0, thin=1,
          ntemps=1, pos=None, reuse_sampler=False, workers=1,
          float_behavior='posterior', is_weighted=True, seed=None,
          progress=True, run_mcmc_kwargs={}):
    r"""Bayesian sampling of the posterior distribution.

    The method uses the ``emcee`` Markov Chain Monte Carlo package and
    assumes that the prior is Uniform. You need to have ``emcee``
    version 3 or newer installed to use this method.

    Parameters
    ----------
    minimizer : minimizer
        Minimizer instance
    params : Parameters, optional
        Parameters to use as starting point. If this is not specified
        then the Parameters used to initialize the Minimizer object
        are used.
    steps : int, optional
        How many samples you would like to draw from the posterior
        distribution for each of the walkers?
    nwalkers : int, optional
        Should be set so :math:`nwalkers >> nvarys`, where ``nvarys``
        are the number of parameters being varied during the fit.
        'Walkers are the members of the ensemble. They are almost like
        separate Metropolis-Hastings chains but, of course, the proposal
        distribution for a given walker depends on the positions of all
        the other walkers in the ensemble.' - from the `emcee` webpage.
    burn : int, optional
        Discard this many samples from the start of the sampling regime.
    thin : int, optional
        Only accept 1 in every `thin` samples.
    ntemps : int, deprecated
        ntemps has no effect.
    pos : numpy.ndarray, optional
        Specify the initial positions for the sampler, an ndarray of
        shape ``(nwalkers, nvarys)``. You can also initialise using a
        previous chain of the same `nwalkers` and ``nvarys``. Note that
        ``nvarys`` may be one larger than you expect it to be if your
        ``userfcn`` returns an array and ``is_weighted=False``.
    reuse_sampler : bool, optional
        Set to True if you have already run `emcee` with the
        `Minimizer` instance and want to continue to draw from its
        ``sampler`` (and so retain the chain history). If False, a
        new sampler is created. The keywords `nwalkers`, `pos`, and
        `params` will be ignored when this is set, as they will be set
        by the existing sampler.
        **Important**: the Parameters used to create the sampler must
        not change in-between calls to `emcee`. Alteration of Parameters
        would include changed ``min``, ``max``, ``vary`` and ``expr``
        attributes. This may happen, for example, if you use an altered
        Parameters object and call the `minimize` method in-between
        calls to `emcee`.
    workers : Pool-like or int, optional
        For parallelization of sampling. It can be any Pool-like object
        with a map method that follows the same calling sequence as the
        built-in `map` function. If int is given as the argument, then
        a multiprocessing-based pool is spawned internally with the
        corresponding number of parallel processes. 'mpi4py'-based
        parallelization and 'joblib'-based parallelization pools can
        also be used here. **Note**: because of multiprocessing
        overhead it may only be worth parallelising if the objective
        function is expensive to calculate, or if there are a large
        number of objective evaluations per step
        (``nwalkers * nvarys``).
    float_behavior : str, optional
        Meaning of float (scalar) output of objective function. Use
        `'posterior'` if it returns a log-posterior probability or
        `'chi2'` if it returns :math:`\chi^2`. See Notes for further
        details.
    is_weighted : bool, optional
        Has your objective function been weighted by measurement
        uncertainties? If ``is_weighted=True`` then your objective
        function is assumed to return residuals that have been divided
        by the true measurement uncertainty ``(data - model) / sigma``.
        If ``is_weighted=False`` then the objective function is
        assumed to return unweighted residuals, ``data - model``. In
        this case `emcee` will employ a positive measurement
        uncertainty during the sampling. This measurement uncertainty
        will be present in the output params and output chain with the
        name ``__lnsigma``. A side effect of this is that you cannot
        use this parameter name yourself.
        **Important**: this parameter only has any effect if your
        objective function returns an array. If your objective function
        returns a float, then this parameter is ignored. See Notes for
        more details.
    seed : int or numpy.random.RandomState, optional
        If `seed` is an ``int``, a new `numpy.random.RandomState`
        instance is used, seeded with `seed`.
        If `seed` is already a `numpy.random.RandomState` instance,
        then that `numpy.random.RandomState` instance is used. Specify
        `seed` for repeatable minimizations.
    progress : bool, optional
        Print a progress bar to the console while running.
    run_mcmc_kwargs : dict, optional
        Additional (optional) keyword arguments that are passed to
        ``emcee.EnsembleSampler.run_mcmc``.

    Returns
    -------
    MinimizerResult
        MinimizerResult object containing updated params, statistics,
        etc. The updated params represent the median of the samples,
        while the uncertainties are half the difference of the 15.87
        and 84.13 percentiles. The `MinimizerResult` contains a few
        additional attributes: `chain` contain the samples and has
        shape ``((steps - burn) // thin, nwalkers, nvarys)``.
        `flatchain` is a `pandas.DataFrame` of the flattened chain,
        that can be accessed with `result.flatchain[parname]`.
        `lnprob` contains the log probability for each sample in
        `chain`. The sample with the highest probability corresponds
        to the maximum likelihood estimate. `acor` is an array
        containing the auto-correlation time for each parameter if the
        auto-correlation time can be computed from the chain. Finally,
        `acceptance_fraction` (an array of the fraction of steps
        accepted for each walker).

    Notes
    -----
    This method samples the posterior distribution of the parameters
    using Markov Chain Monte Carlo. It calculates the log-posterior
    probability of the model parameters, `F`, given the data, `D`,
    :math:`\ln p(F_{true} | D)`. This 'posterior probability' is
    given by:

    .. math::

        \ln p(F_{true} | D) \propto \ln p(D | F_{true}) + \ln p(F_{true})

    where :math:`\ln p(D | F_{true})` is the 'log-likelihood' and
    :math:`\ln p(F_{true})` is the 'log-prior'. The default log-prior
    encodes prior information known about the model that the log-prior
    probability is ``-numpy.inf`` (impossible) if any of the parameters
    is outside its limits, and is zero if all the parameters are inside
    their bounds (uniform prior). The log-likelihood function is [1]_:

    .. math::

        \ln p(D|F_{true}) = -\frac{1}{2}\sum_n \left[\frac{(g_n(F_{true}) - D_n)^2}{s_n^2}+\ln (2\pi s_n^2)\right]

    The first term represents the residual (:math:`g` being the
    generative model, :math:`D_n` the data and :math:`s_n` the
    measurement uncertainty). This gives :math:`\chi^2` when summed
    over all data points. The objective function may also return the
    log-posterior probability, :math:`\ln p(F_{true} | D)`. Since the
    default log-prior term is zero, the objective function can also
    just return the log-likelihood, unless you wish to create a
    non-uniform prior.

    If the objective function returns a float value, this is assumed
    by default to be the log-posterior probability, (`float_behavior`
    default is 'posterior'). If your objective function returns
    :math:`\chi^2`, then you should use ``float_behavior='chi2'``
    instead.

    By default objective functions may return an ndarray of (possibly
    weighted) residuals. In this case, use `is_weighted` to select
    whether these are correctly weighted by measurement uncertainty.
    Note that this ignores the second term above, so that to calculate
    a correct log-posterior probability value your objective function
    should return a float value. With ``is_weighted=False`` the data
    uncertainty, `s_n`, will be treated as a nuisance parameter to be
    marginalized out. This uses strictly positive uncertainty
    (homoscedasticity) for each data point,
    :math:`s_n = \exp(\rm{\_\_lnsigma})`. ``__lnsigma`` will be
    present in `MinimizerResult.params`, as well as `Minimizer.chain`
    and ``nvarys`` will be increased by one.

    References
    ----------
    .. [1] https://emcee.readthedocs.io

    """
    if not HAS_EMCEE:
        raise NotImplementedError('emcee version 3 is required.')

    if ntemps > 1:
        msg = ("'ntemps' has no effect anymore, since the PTSampler was "
               "removed from emcee version 3.")
        raise DeprecationWarning(msg)

    tparams = params
    # if you're reusing the sampler then nwalkers have to be
    # determined from the previous sampling
    if reuse_sampler:
        if not hasattr(self, 'sampler') or not hasattr(self, '_lastpos'):
            raise ValueError("You wanted to use an existing sampler, but "
                             "it hasn't been created yet")
        if len(self._lastpos.shape) == 2:
            nwalkers = self._lastpos.shape[0]
        elif len(self._lastpos.shape) == 3:
            nwalkers = self._lastpos.shape[1]
        tparams = None

    result = self.prepare_fit(params=tparams)
    params = result.params

    # check if the userfcn returns a vector of residuals
    out = self.userfcn(params, *self.userargs, **self.userkws)
    out = np.asarray(out).ravel()
    if out.size > 1 and is_weighted is False and '__lnsigma' not in params:
        # __lnsigma should already be in params if is_weighted was
        # previously set to True.
        params.add('__lnsigma', value=0.01, min=-np.inf, max=np.inf,
                   vary=True)
        # have to re-prepare the fit
        result = self.prepare_fit(params)
        params = result.params

    result.method = 'emcee'

    # Removing internal parameter scaling. We could possibly keep it,
    # but I don't know how this affects the emcee sampling.
    bounds = []
    var_arr = np.zeros(len(result.var_names))
    i = 0
    for par in params:
        param = params[par]
        if param.expr is not None:
            param.vary = False
        if param.vary:
            var_arr[i] = param.value
            i += 1
        else:
            # don't want to append bounds if they're not being varied.
            continue
        param.from_internal = lambda val: val
        lb, ub = param.min, param.max
        if lb is None or lb is np.nan:
            lb = -np.inf
        if ub is None or ub is np.nan:
            ub = np.inf
        bounds.append((lb, ub))
    bounds = np.array(bounds)

    self.nvarys = len(result.var_names)

    # set up multiprocessing options for the samplers
    auto_pool = None
    sampler_kwargs = {}
    if isinstance(workers, int) and workers > 1:
        auto_pool = multiprocessing.Pool(workers)
        sampler_kwargs['pool'] = auto_pool
    elif hasattr(workers, 'map'):
        sampler_kwargs['pool'] = workers

    # function arguments for the log-probability functions
    # these values are sent to the log-probability functions by the sampler.
    lnprob_args = (self.userfcn, params, result.var_names, bounds)
    lnprob_kwargs = {'is_weighted': is_weighted,
                     'float_behavior': float_behavior,
                     'userargs': self.userargs,
                     'userkws': self.userkws,
                     'nan_policy': self.nan_policy}

    sampler_kwargs['args'] = lnprob_args
    sampler_kwargs['kwargs'] = lnprob_kwargs

    # set up the random number generator
    rng = _make_random_gen(seed)

    # now initialise the samplers
    if reuse_sampler:
        if auto_pool is not None:
            self.sampler.pool = auto_pool

        p0 = self._lastpos
        if p0.shape[-1] != self.nvarys:
            raise ValueError("You cannot reuse the sampler if the number "
                             "of varying parameters has changed")

    else:
        p0 = 1 + rng.randn(nwalkers, self.nvarys) * 1.e-4
        p0 *= var_arr
        sampler_kwargs.setdefault('pool', auto_pool)
        self.sampler = emcee.EnsembleSampler(nwalkers, self.nvarys,
                                             self._lnprob, **sampler_kwargs)

    # user supplies an initialisation position for the chain
    # If you try to run the sampler with p0 of a wrong size then you'll get
    # a ValueError. Note, you can't initialise with a position if you are
    # reusing the sampler.
    if pos is not None and not reuse_sampler:
        tpos = np.asfarray(pos)
        if p0.shape == tpos.shape:
            pass
        # trying to initialise with a previous chain
        elif tpos.shape[-1] == self.nvarys:
            tpos = tpos[-1]
        else:
            raise ValueError('pos should have shape (nwalkers, nvarys)')
        p0 = tpos

    # if you specified a seed then you also need to seed the sampler
    if seed is not None:
        self.sampler.random_state = rng.get_state()

    if not isinstance(run_mcmc_kwargs, dict):
        raise ValueError('run_mcmc_kwargs should be a dict of keyword arguments')

    # now do a production run, sampling all the time
    try:
        output = self.sampler.run_mcmc(p0, steps, progress=progress, **run_mcmc_kwargs)
        self._lastpos = output.coords
    except AbortFitException:
        result.aborted = True
        result.message = "Fit aborted by user callback. Could not estimate error-bars."
        result.success = False
        result.nfev = self.result.nfev

    # discard the burn samples and thin
    chain = self.sampler.get_chain(thin=thin, discard=burn)[..., :, :]
    lnprobability = self.sampler.get_log_prob(thin=thin, discard=burn)[..., :]
    flatchain = chain.reshape((-1, self.nvarys))
    if not result.aborted:
        quantiles = np.percentile(flatchain, [15.87, 50, 84.13], axis=0)

        for i, var_name in enumerate(result.var_names):
            std_l, median, std_u = quantiles[:, i]
            params[var_name].value = median
            params[var_name].stderr = 0.5 * (std_u - std_l)
            params[var_name].correl = {}

        params.update_constraints()

        # work out correlation coefficients
        corrcoefs = np.corrcoef(flatchain.T)

        for i, var_name in enumerate(result.var_names):
            for j, var_name2 in enumerate(result.var_names):
                if i != j:
                    result.params[var_name].correl[var_name2] = corrcoefs[i, j]

    result.chain = np.copy(chain)
    result.lnprob = np.copy(lnprobability)
    result.errorbars = True
    result.nvarys = len(result.var_names)
    result.nfev = nwalkers*steps

    try:
        result.acor = self.sampler.get_autocorr_time()
    except AutocorrError as e:
        print(str(e))
    result.acceptance_fraction = self.sampler.acceptance_fraction

    # Calculate the residual with the "best fit" parameters
    out = self.userfcn(params, *self.userargs, **self.userkws)
    result.residual = _nan_policy(out, nan_policy=self.nan_policy,
                                  handle_inf=False)

    # If uncertainty was automatically estimated, weight the residual properly
    if not is_weighted and result.residual.size > 1 and '__lnsigma' in params:
        result.residual /= np.exp(params['__lnsigma'].value)

    # Calculate statistics for the two standard cases:
    if isinstance(result.residual, np.ndarray) or (float_behavior == 'chi2'):
        result._calculate_statistics()

    # Handle special case unique to emcee:
    # This should eventually be moved into result._calculate_statistics.
    elif float_behavior == 'posterior':
        result.ndata = 1
        result.nfree = 1

        # assuming prior prob = 1, this is true
        _neg2_log_likel = -2*result.residual

        # assumes that residual is properly weighted, avoid overflowing np.exp()
        result.chisqr = np.exp(min(650, _neg2_log_likel))

        result.redchi = result.chisqr / result.nfree
        result.aic = _neg2_log_likel + 2 * result.nvarys
        result.bic = _neg2_log_likel + np.log(result.ndata) * result.nvarys

    if auto_pool is not None:
        auto_pool.terminate()

    return result



from benchmarks.benchmarks import MinimizerClassSuite, MinimizeSuite

mtest = MinimizeSuite()
mtest.setup()
out = mtest.time_minimize()
out = mtest.time_minimize_large()
out = mtest.time_minimize_withnan()
out = mtest.time_confinterval()

mtest = MinimizerClassSuite()
mtest.setup()
out = mtest.time_differential_evolution()
out = mtest.time_emcee()



"""Tests for the Model, CompositeModel, and ModelResult classes."""

import functools
import unittest
import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest
from scipy import __version__ as scipy_version

import lmfit
from lmfit import Model, Parameters, models
from lmfit.lineshapes import gaussian, lorentzian, step, voigt
from lmfit.model import get_reducer, propagate_err
from lmfit.models import GaussianModel, PseudoVoigtModel


@pytest.fixture()
def gmodel():
    """Return a Gaussian model."""
    return Model(gaussian)


def test_get_reducer_invalid_option():
    """Tests for ValueError when using an unsupported option."""
    option = 'unknown'
    msg = r'Invalid option'
    with pytest.raises(ValueError, match=msg):
        get_reducer(option)


test_data_get_reducer = [('real', [1.0, 1.0, 2.0, 2.0]),
                         ('imag', [0.0, 10.0, 0.0, 20.0]),
                         ('abs', [1.0, 10.04987562, 2.0, 20.09975124]),
                         ('angle', [0.0, 1.47112767, 0.0, 1.471127670])]


@pytest.mark.parametrize('option, expected_array', test_data_get_reducer)
def test_get_reducer(option, expected_array):
    """Tests for ValueError when using an unsupported option."""
    complex_array = np.array([1.0, 1.0+10j, 2.0, 2.0+20j], dtype='complex')
    func = get_reducer(option)
    real_array = func(complex_array)

    assert np.all(np.isreal(real_array))
    assert_allclose(real_array, expected_array)

    # nothing should happen to an array that only contains real data
    assert_allclose(func(real_array), real_array)


def test_propagate_err_invalid_option():
    """Tests for ValueError when using an unsupported option."""
    z = np.array([0, 1, 2, 3, 4, 5])
    dz = np.random.normal(size=z.size, scale=0.1)
    option = 'unknown'
    msg = r'Invalid option'
    with pytest.raises(ValueError, match=msg):
        propagate_err(z, dz, option)


def test_propagate_err_unequal_shape_z_dz():
    """Tests for ValueError when using unequal arrays for z and dz."""
    z = np.array([0, 1, 2, 3, 4, 5])
    dz = np.random.normal(size=z.size-1, scale=0.1)
    msg = r'shape of z:'
    with pytest.raises(ValueError, match=msg):
        propagate_err(z, dz, option='abs')


@pytest.mark.parametrize('option', ['real', 'imag', 'abs', 'angle'])
def test_propagate_err(option):
    """Tests for ValueError when using an unsupported option."""
    np.random.seed(2020)
    z = np.array([1.0, 1.0+10j, 2.0, 2.0+20j], dtype='complex')
    dz = np.random.normal(z.size, scale=0.1)*z

    # if `z` is real, assume that `dz` is also real and return it as-is
    err = propagate_err(np.real(z), np.real(dz), option)
    assert_allclose(err, np.real(dz))

    # if `z` is complex, but `dz` is real apply the err to both real/imag
    err_complex_real = propagate_err(z, np.real(dz), option)
    assert np.all(np.isreal(err_complex_real))
    dz_used = np.real(dz)+1j*np.real(dz)
    if option == 'real':
        assert_allclose(err_complex_real, np.real(dz_used))
    elif option == 'imag':
        assert_allclose(err_complex_real, np.imag(dz_used))
    elif option == 'abs':
        assert_allclose(err_complex_real,
                        [3.823115, 3.823115, 7.646231, 7.646231],
                        rtol=1.0e-5)
    elif option == 'angle':
        assert_allclose(err_complex_real,
                        [3.823115, 0.380414, 3.823115, 0.380414],
                        rtol=1.0e-5)

    # both `z` and `dz` are complex
    err_complex_complex = propagate_err(z, dz, option)
    assert np.all(np.isreal(err_complex_complex))
    if option == 'real':
        assert_allclose(err_complex_complex, np.real(dz))
    elif option == 'imag':
        assert_allclose(err_complex_complex, np.imag(dz))
    elif option == 'abs':
        assert_allclose(err_complex_complex,
                        [3.823115, 38.043322, 7.646231, 76.086645],
                        rtol=1.0e-5)
    elif option == 'angle':
        assert_allclose(err_complex_complex, [0., 0.535317, 0., 0.535317],
                        rtol=1.0e-5)


def test_initialize_Model_class_default_arguments(gmodel):
    """Test for Model class initialized with default arguments."""
    assert gmodel.prefix == ''
    assert gmodel._param_root_names == ['amplitude', 'center', 'sigma']
    assert gmodel.param_names == ['amplitude', 'center', 'sigma']
    assert gmodel.independent_vars == ['x']
    assert gmodel.nan_policy == 'raise'
    assert gmodel.name == 'Model(gaussian)'
    assert gmodel.opts == {}
    assert gmodel.def_vals == {'amplitude': 1.0, 'center': 0.0, 'sigma': 1.0}


def test_initialize_Model_class_independent_vars():
    """Test for Model class initialized with independent_vars."""
    model = Model(gaussian, independent_vars=['amplitude'])
    assert model._param_root_names == ['x', 'center', 'sigma']
    assert model.param_names == ['x', 'center', 'sigma']
    assert model.independent_vars == ['amplitude']


def test_initialize_Model_class_param_names():
    """Test for Model class initialized with param_names."""
    model = Model(gaussian, param_names=['amplitude'])

    assert model._param_root_names == ['amplitude']
    assert model.param_names == ['amplitude']


@pytest.mark.parametrize("policy", ['raise', 'omit', 'propagate'])
def test_initialize_Model_class_nan_policy(policy):
    """Test for Model class initialized with nan_policy."""
    model = Model(gaussian, nan_policy=policy)

    assert model.nan_policy == policy


def test_initialize_Model_class_prefix():
    """Test for Model class initialized with prefix."""
    model = Model(gaussian, prefix='test_')

    assert model.prefix == 'test_'
    assert model._param_root_names == ['amplitude', 'center', 'sigma']
    assert model.param_names == ['test_amplitude', 'test_center', 'test_sigma']
    assert model.name == "Model(gaussian, prefix='test_')"

    model = Model(gaussian, prefix=None)

    assert model.prefix == ''


def test_initialize_Model_name():
    """Test for Model class initialized with name."""
    model = Model(gaussian, name='test_function')

    assert model.name == 'Model(test_function)'


def test_initialize_Model_kws():
    """Test for Model class initialized with **kws."""
    kws = {'amplitude': 10.0}
    model = Model(gaussian,
                  independent_vars=['x', 'amplitude'], **kws)

    assert model._param_root_names == ['center', 'sigma']
    assert model.param_names == ['center', 'sigma']
    assert model.independent_vars == ['x', 'amplitude']
    assert model.opts == kws


test_reprstring_data = [(False, 'Model(gaussian)'),
                        (True, "Model(gaussian, amplitude='10.0')")]


@pytest.mark.parametrize("option, expected", test_reprstring_data)
def test_Model_reprstring(option, expected):
    """Test for Model class function _reprstring."""
    kws = {'amplitude': 10.0}
    model = Model(gaussian,
                  independent_vars=['x', 'amplitude'], **kws)

    assert model._reprstring(option) == expected


def test_Model_get_state(gmodel):
    """Test for Model class function _get_state."""
    out = gmodel._get_state()

    assert isinstance(out, tuple)
    assert out[1] == out[2] is None
    sstate = out[0]
    if isinstance(sstate, tuple):  # state version 1
        assert sstate[1] is not None
        assert sstate[0] == 'gaussian'
        assert sstate[2:] == ('gaussian', '', ['x'],
                              ['amplitude', 'center', 'sigma'], {}, 'raise', {})
    elif isinstance(sstate, dict) and 'version' in sstate:  # state version 2 or higher
        if sstate['version'] == '2':
            assert sstate['funcname'] == 'gaussian'
            assert sstate['name'] == 'gaussian'
            assert sstate['independent_vars'] == ['x']
            assert sstate['param_root_names'] == ['amplitude', 'center', 'sigma']
        else:
            assert sstate['version'] == 'unknown version!'
    else:
        assert sstate == 'unknown model state'


def test_Model_set_state(gmodel):
    """Test for Model class function _set_state.

    This function is just calling `_buildmodel`, which will be tested
    below together with the use of `funcdefs`.

    """
    out = gmodel._get_state()

    new_model = Model(lorentzian)
    new_model = new_model._set_state(out)

    assert new_model.prefix == gmodel.prefix
    assert new_model._param_root_names == gmodel._param_root_names
    assert new_model.param_names == gmodel.param_names
    assert new_model.independent_vars == gmodel.independent_vars
    assert new_model.nan_policy == gmodel.nan_policy
    assert new_model.name == gmodel.name
    assert new_model.opts == gmodel.opts


def test_Model_dumps_loads(gmodel):
    """Test for Model class functions dumps and loads.

    These function are used when saving/loading the Model class and will be
    tested more thoroughly in test_model_saveload.py.

    """
    model_json = gmodel.dumps()
    _ = gmodel.loads(model_json)


def test_Model_getter_setter_name(gmodel):
    """Test for Model class getter/setter functions for name."""
    assert gmodel.name == 'Model(gaussian)'

    gmodel.name = 'test_gaussian'
    assert gmodel.name == 'Model(test_gaussian)'


def test_Model_getter_setter_prefix(gmodel):
    """Test for Model class getter/setter functions for prefix."""
    assert gmodel.prefix == ''
    assert gmodel.param_names == ['amplitude', 'center', 'sigma']

    gmodel.prefix = 'g1_'
    assert gmodel.prefix == 'g1_'
    assert gmodel.param_names == ['g1_amplitude', 'g1_center', 'g1_sigma']

    gmodel.prefix = ''
    assert gmodel.prefix == ''
    assert gmodel.param_names == ['amplitude', 'center', 'sigma']


def test_Model_getter_param_names(gmodel):
    """Test for Model class getter function for param_names."""
    assert gmodel.param_names == ['amplitude', 'center', 'sigma']


def test_Model__repr__(gmodel):
    """Test for Model class __repr__ method."""
    assert gmodel.__repr__() == '<lmfit.Model: Model(gaussian)>'


def test_Model_copy(gmodel):
    """Test for Model class copy method."""
    msg = 'Model.copy does not work. Make a new Model'
    with pytest.raises(NotImplementedError, match=msg):
        gmodel.copy()


def test__parse_params_func_None():
    """Test for _parse_params function with func=None."""
    mod = Model(None)

    assert mod._prefix == ''
    assert mod.func is None
    assert mod._func_allargs == []
    assert mod._func_haskeywords is False
    assert mod.independent_vars == []


def test__parse_params_asteval_functions():
    """Test for _parse_params function with asteval functions."""
    # TODO: cannot find a use-case for this....
    pass


def test__parse_params_inspect_signature():
    """Test for _parse_params function using inspect.signature."""
    # 1. function with a positional argument
    def func_var_positional(a, *b):
        pass

    with pytest.raises(ValueError, match=r"varargs '\*b' is not supported"):
        Model(func_var_positional)

    # 2. function with a keyword argument
    def func_keyword(a, b, **c):
        pass

    mod = Model(func_keyword)
    assert mod._func_allargs == ['a', 'b']
    assert mod._func_haskeywords is True
    assert mod.independent_vars == ['a']
    assert mod.def_vals == {}

    # 3. function with keyword argument only
    def func_keyword_only(**b):
        pass

    mod = Model(func_keyword_only)
    assert mod._func_allargs == []
    assert mod._func_haskeywords is True
    assert mod.independent_vars == []
    assert mod._param_root_names is None

    # 4. function with default value
    def func_default_value(a, b, c=10):
        pass

    mod = Model(func_default_value)
    assert mod._func_allargs == ['a', 'b', 'c']
    assert mod._func_haskeywords is False
    assert mod.independent_vars == ['a']

    assert isinstance(mod.def_vals, dict)
    assert_allclose(mod.def_vals['c'], 10)


def test_make_params_withprefixs():
    # tests Github Issue #893
    gmod1 = GaussianModel(prefix='p1_')
    gmod2 = GaussianModel(prefix='p2_')

    model = gmod1 + gmod2

    pars_1a = gmod1.make_params(p1_amplitude=10, p1_center=600, p1_sigma=3)
    pars_1b = gmod1.make_params(amplitude=10, center=600, sigma=3)

    pars_2a = gmod2.make_params(p2_amplitude=30, p2_center=730, p2_sigma=4)
    pars_2b = gmod2.make_params(amplitude=30, center=730, sigma=4)

    pars_a = Parameters()
    pars_a.update(pars_1a)
    pars_a.update(pars_2a)

    pars_b = Parameters()
    pars_b.update(pars_1b)
    pars_b.update(pars_2b)

    pars_c = model.make_params()

    for pname in ('p1_amplitude', 'p1_center', 'p1_sigma',
                  'p2_amplitude', 'p2_center', 'p2_sigma'):
        assert pname in pars_a
        assert pname in pars_b
        assert pname in pars_c


def test__parse_params_forbidden_variable_names():
    """Tests for _parse_params function using invalid variable names."""

    def func_invalid_var(data, a):
        pass

    def func_invalid_par(a, weights):
        pass

    msg = r"Invalid independent variable name \('data'\) for function func_invalid_var"
    with pytest.raises(ValueError, match=msg):
        Model(func_invalid_var)

    msg = r"Invalid parameter name \('weights'\) for function func_invalid_par"
    with pytest.raises(ValueError, match=msg):
        Model(func_invalid_par)


@pytest.mark.parametrize('input_dtype', (np.int16, np.int32, np.float32,
                                         np.complex64, np.complex128, 'list',
                                         'tuple', 'pandas-real',
                                         'pandas-complex'))
def test_coercion_of_input_data(peakdata, input_dtype):
    """Test for coercion of 'data' and 'independent_vars'.

    'data' and `independent_vars` should be coerced to 'float64' or 'complex128'

    unless told not be coerced by setting ``coerce_farray=False``.

    # - dtype for 'indepdendent_vars' is only changed when the input is a list,
    #    tuple, numpy.ndarray, or pandas.Series

    """
    x, y = peakdata

    def gaussian_lists(x, amplitude=1.0, center=0.0, sigma=1.0):
        xarr = np.array(x, dtype=np.float64)
        return ((amplitude/(max(1.e-15, np.sqrt(2*np.pi)*sigma)))
                * np.exp(-(xarr-center)**2 / max(1.e-15, (2*sigma**2))))

    for coerce_farray in True, False:
        if (input_dtype in ('pandas-real', 'pandas-complex')
           and not lmfit.minimizer.HAS_PANDAS):
            return

        if not coerce_farray and input_dtype in ('list', 'tuple'):
            model = lmfit.Model(gaussian_lists)
        else:
            model = lmfit.Model(gaussian)

        pars = model.make_params(amplitude=5, center=10, sigma=2)

        if input_dtype == 'pandas-real':
            result = model.fit(lmfit.model.Series(y, dtype=np.float32), pars,
                               x=lmfit.model.Series(x, dtype=np.float32),
                               coerce_farray=coerce_farray)

            expected_dtype = np.float64 if coerce_farray else np.float32

        elif input_dtype == 'pandas-complex':
            result = model.fit(lmfit.model.Series(y, dtype=np.complex64), pars,
                               x=lmfit.model.Series(x, dtype=np.complex64),
                               coerce_farray=coerce_farray)
            expected_dtype = np.complex128 if coerce_farray else np.complex64

        elif input_dtype == 'list':
            result = model.fit(y.tolist(), pars, x=x.tolist(),
                               coerce_farray=coerce_farray)
            expected_dtype = np.float64 if coerce_farray else list

        elif input_dtype == 'tuple':
            result = model.fit(tuple(y), pars, x=tuple(x),
                               coerce_farray=coerce_farray)
            expected_dtype = np.float64 if coerce_farray else tuple

        else:
            result = model.fit(np.asarray(y, dtype=input_dtype), pars,
                               x=np.asarray(x, dtype=input_dtype),
                               coerce_farray=coerce_farray)
            expected_dtype = np.float64
            if input_dtype in (np.complex64, np.complex128):
                expected_dtype = np.complex128
            expected_dtype = expected_dtype if coerce_farray else input_dtype

        if not coerce_farray and input_dtype in ('list', 'tuple'):
            assert isinstance(result.userkws['x'], (list, tuple))
            assert isinstance(result.userargs[0], (list, tuple))
        else:
            assert result.userkws['x'].dtype == expected_dtype
            assert result.userargs[0].dtype == expected_dtype


def test_figure_default_title(peakdata):
    """Test default figure title."""
    pytest.importorskip('matplotlib')

    x, y = peakdata
    pvmodel = PseudoVoigtModel()
    params = pvmodel.guess(y, x=x)
    result = pvmodel.fit(y, params, x=x)

    ax = result.plot_fit()
    assert ax.axes.get_title() == 'Model(pvoigt)'

    ax = result.plot_residuals()
    assert ax.axes.get_title() == 'Model(pvoigt)'

    fig = result.plot()
    assert fig.axes[0].get_title() == 'Model(pvoigt)'  # default model.name
    assert fig.axes[1].get_title() == ''  # no title for fit subplot


def test_figure_title_using_title_keyword_argument(peakdata):
    """Test setting figure title using title keyword argument."""
    pytest.importorskip('matplotlib')

    x, y = peakdata
    pvmodel = PseudoVoigtModel()
    params = pvmodel.guess(y, x=x)
    result = pvmodel.fit(y, params, x=x)

    ax = result.plot_fit(title='test')
    assert ax.axes.get_title() == 'test'

    ax = result.plot_residuals(title='test')
    assert ax.axes.get_title() == 'test'

    fig = result.plot(title='test')
    assert fig.axes[0].get_title() == 'test'
    assert fig.axes[1].get_title() == ''  # no title for fit subplot


def test_figure_title_using_title_to_ax_kws(peakdata):
    """Test setting figure title by supplying ax_{fit,res}_kws."""
    pytest.importorskip('matplotlib')

    x, y = peakdata
    pvmodel = PseudoVoigtModel()
    params = pvmodel.guess(y, x=x)
    result = pvmodel.fit(y, params, x=x)

    ax = result.plot_fit(ax_kws={'title': 'ax_kws'})
    assert ax.axes.get_title() == 'ax_kws'

    ax = result.plot_residuals(ax_kws={'title': 'ax_kws'})
    assert ax.axes.get_title() == 'ax_kws'

    fig = result.plot(ax_res_kws={'title': 'ax_res_kws'})
    assert fig.axes[0].get_title() == 'ax_res_kws'
    assert fig.axes[1].get_title() == ''

    fig = result.plot(ax_fit_kws={'title': 'ax_fit_kws'})
    assert fig.axes[0].get_title() == 'Model(pvoigt)'  # default model.name
    assert fig.axes[1].get_title() == ''  # no title for fit subplot


def test_priority_setting_figure_title(peakdata):
    """Test for setting figure title: title keyword argument has priority."""
    pytest.importorskip('matplotlib')

    x, y = peakdata
    pvmodel = PseudoVoigtModel()
    params = pvmodel.guess(y, x=x)
    result = pvmodel.fit(y, params, x=x)

    ax = result.plot_fit(ax_kws={'title': 'ax_kws'}, title='test')
    assert ax.axes.get_title() == 'test'

    ax = result.plot_residuals(ax_kws={'title': 'ax_kws'}, title='test')
    assert ax.axes.get_title() == 'test'

    fig = result.plot(ax_res_kws={'title': 'ax_res_kws'}, title='test')
    assert fig.axes[0].get_title() == 'test'
    assert fig.axes[1].get_title() == ''

    fig = result.plot(ax_fit_kws={'title': 'ax_fit_kws'}, title='test')
    assert fig.axes[0].get_title() == 'test'
    assert fig.axes[1].get_title() == ''


def test_eval_with_kwargs():
    # Check eval() with both params and kwargs, even when there are
    # constraints
    x = np.linspace(0, 30, 301)
    np.random.seed(13)
    y1 = (gaussian(x, amplitude=10, center=12.0, sigma=2.5) +
          gaussian(x, amplitude=20, center=19.0, sigma=2.5))

    y2 = (gaussian(x, amplitude=10, center=12.0, sigma=1.5) +
          gaussian(x, amplitude=20, center=19.0, sigma=2.5))

    model = Model(gaussian, prefix='g1_') + Model(gaussian, prefix='g2_')
    params = model.make_params(g1_amplitude=10, g1_center=12.0, g1_sigma=1,
                               g2_amplitude=20, g2_center=19.0,
                               g2_sigma={'expr': 'g1_sigma'})

    r1 = model.eval(params, g1_sigma=2.5, x=x)
    assert_allclose(r1, y1, atol=1.e-3)

    assert params['g2_sigma'].value == 1
    assert params['g1_sigma'].value == 1

    params['g1_sigma'].value = 1.5
    params['g2_sigma'].expr = None
    params['g2_sigma'].value = 2.5

    r2 = model.eval(params, x=x)
    assert_allclose(r2, y2, atol=1.e-3)


def test_guess_requires_x():
    """Test to make sure that ``guess()`` method requires the argument ``x``.

    The ``guess`` method needs ``x`` values (i.e., the independent variable)
    to estimate initial parameters, but this was not a required argument.
    See GH #747.

    """
    mod = lmfit.model.Model(gaussian)

    msg = r"guess\(\) missing 2 required positional arguments: 'data' and 'x'"
    with pytest.raises(TypeError, match=msg):
        mod.guess()


# Below is the content of the original test_model.py file. These tests still
# need to be checked and possibly updated to the pytest-style. They work fine
# though so leave them in for now.

def assert_results_close(actual, desired, rtol=1e-03, atol=1e-03, err_msg='',
                         verbose=True):
    for param_name, value in desired.items():
        assert_allclose(actual[param_name], value, rtol, atol, err_msg,
                        verbose)


def firstarg_ndarray(func):
    """a simple wrapper used for testing that wrapped
    functions can be model functions"""
    @functools.wraps(func)
    def wrapper(x, *args, **kws):
        x = np.asarray(x)
        return func(x, *args, **kws)
    return wrapper


@firstarg_ndarray
def linear_func(x, a, b):
    "test wrapped model function"
    return a*x+b


class CommonTests:
    # to be subclassed for testing predefined models

    def setUp(self):
        np.random.seed(1)
        self.noise = 0.0001*np.random.randn(self.x.size)
        # Some Models need args (e.g., polynomial order), and others don't.
        try:
            args = self.args
        except AttributeError:
            self.model = self.model_constructor()
            self.model_omit = self.model_constructor(nan_policy='omit')
            self.model_raise = self.model_constructor(nan_policy='raise')
            self.model_explicit_var = self.model_constructor(['x'])
            func = self.model.func
        else:
            self.model = self.model_constructor(*args)
            self.model_omit = self.model_constructor(*args, nan_policy='omit')
            self.model_raise = self.model_constructor(*args, nan_policy='raise')
            self.model_explicit_var = self.model_constructor(
                *args, independent_vars=['x'])
            func = self.model.func
        self.data = func(x=self.x, **self.true_values()) + self.noise

    @property
    def x(self):
        return np.linspace(1, 10, num=1000)

    def test_fit(self):
        model = self.model

        # Pass Parameters object.
        params = model.make_params(**self.guess())
        result = model.fit(self.data, params, x=self.x)
        assert_results_close(result.values, self.true_values())

        # Pass individual Parameter objects as kwargs.
        kwargs = dict(params.items())
        result = self.model.fit(self.data, x=self.x, **kwargs)
        assert_results_close(result.values, self.true_values())

        # Pass guess values (not Parameter objects) as kwargs.
        kwargs = {name: p.value for name, p in params.items()}
        result = self.model.fit(self.data, x=self.x, **kwargs)
        assert_results_close(result.values, self.true_values())

    def test_explicit_independent_vars(self):
        self.check_skip_independent_vars()
        model = self.model_explicit_var
        pars = model.make_params(**self.guess())
        result = model.fit(self.data, pars, x=self.x)
        assert_results_close(result.values, self.true_values())

    def test_fit_with_weights(self):
        model = self.model

        # fit without weights
        params = model.make_params(**self.guess())
        out1 = model.fit(self.data, params, x=self.x)

        # fit with weights
        weights = 1.0/(0.5 + self.x**2)
        out2 = model.fit(self.data, params, weights=weights, x=self.x)

        max_diff = 0.0
        for parname, val1 in out1.values.items():
            val2 = out2.values[parname]
            if max_diff < abs(val1-val2):
                max_diff = abs(val1-val2)
        assert max_diff > 1.e-8

    def test_result_attributes(self):
        pars = self.model.make_params(**self.guess())
        result = self.model.fit(self.data, pars, x=self.x)

        # result.init_values
        assert_results_close(result.values, self.true_values())
        self.assertEqual(result.init_values, self.guess())

        # result.init_params
        params = self.model.make_params()
        for param_name, value in self.guess().items():
            params[param_name].value = value
        self.assertEqual(result.init_params, params)

        # result.best_fit
        assert_allclose(result.best_fit, self.data, atol=self.noise.max())

        # result.init_fit
        init_fit = self.model.func(x=self.x, **self.guess())
        assert_allclose(result.init_fit, init_fit)

        # result.model
        self.assertTrue(result.model is self.model)

    def test_result_eval(self):
        # Check eval() output against init_fit and best_fit.
        pars = self.model.make_params(**self.guess())
        result = self.model.fit(self.data, pars, x=self.x)
        assert_allclose(result.eval(x=self.x, **result.values),
                        result.best_fit)
        assert_allclose(result.eval(x=self.x, **result.init_values),
                        result.init_fit)

    def test_result_eval_custom_x(self):
        self.check_skip_independent_vars()
        pars = self.model.make_params(**self.guess())
        result = self.model.fit(self.data, pars, x=self.x)

        # Check that the independent variable is respected.
        short_eval = result.eval(x=np.array([0, 1, 2]), **result.values)
        if hasattr(short_eval, '__len__'):
            self.assertEqual(len(short_eval), 3)

    def test_result_report(self):
        pars = self.model.make_params(**self.guess())
        result = self.model.fit(self.data, pars, x=self.x)
        report = result.fit_report()
        assert "[[Model]]" in report
        assert "[[Variables]]" in report
        assert "[[Fit Statistics]]" in report
        assert " # function evals   =" in report
        assert " Akaike " in report
        assert " chi-square " in report

    def test_data_alignment(self):
        pytest.importorskip('pandas')

        from pandas import Series

        # Align data and indep var of different lengths using pandas index.
        data = Series(self.data.copy()).iloc[10:-10]
        x = Series(self.x.copy())

        model = self.model
        params = model.make_params(**self.guess())
        result = model.fit(data, params, x=x)
        result = model.fit(data, params, x=x)
        assert_results_close(result.values, self.true_values())

        # Skip over missing (NaN) values, aligning via pandas index.
        data.iloc[500:510] = np.nan
        result = self.model_omit.fit(data, params, x=x)
        assert_results_close(result.values, self.true_values())

        # Raise if any NaN values are present.
        raises = lambda: self.model_raise.fit(data, params, x=x)
        self.assertRaises(ValueError, raises)

    def check_skip_independent_vars(self):
        # to be overridden for models that do not accept indep vars
        pass

    def test_aic(self):
        model = self.model

        # Pass Parameters object.
        params = model.make_params(**self.guess())
        result = model.fit(self.data, params, x=self.x)
        aic = result.aic
        self.assertTrue(aic < 0)  # aic must be negative

        # Pass extra unused Parameter.
        params.add("unused_param", value=1.0, vary=True)
        result = model.fit(self.data, params, x=self.x)
        aic_extra = result.aic
        self.assertTrue(aic_extra < 0)  # aic must be negative
        self.assertTrue(aic < aic_extra)  # extra param should lower the aic

    def test_bic(self):
        model = self.model

        # Pass Parameters object.
        params = model.make_params(**self.guess())
        result = model.fit(self.data, params, x=self.x)
        bic = result.bic
        self.assertTrue(bic < 0)  # aic must be negative

        # Compare to AIC
        aic = result.aic
        self.assertTrue(aic < bic)  # aic should be lower than bic

        # Pass extra unused Parameter.
        params.add("unused_param", value=1.0, vary=True)
        result = model.fit(self.data, params, x=self.x)
        bic_extra = result.bic
        self.assertTrue(bic_extra < 0)  # bic must be negative
        self.assertTrue(bic < bic_extra)  # extra param should lower the bic


class TestUserDefiniedModel(CommonTests, unittest.TestCase):
    # mainly aimed at checking that the API does what it says it does
    # and raises the right exceptions or warnings when things are not right
    def setUp(self):
        self.true_values = lambda: dict(amplitude=7.1, center=1.1, sigma=2.40)
        self.guess = lambda: dict(amplitude=5, center=2, sigma=4)
        # return a fresh copy
        self.model_constructor = (
            lambda *args, **kwargs: Model(gaussian, *args, **kwargs))
        super().setUp()

    @property
    def x(self):
        return np.linspace(-10, 10, num=1000)

    def test_lists_become_arrays(self):
        # smoke test
        self.model.fit([1, 2, 3], x=[1, 2, 3], **self.guess())
        pytest.raises(ValueError,
                      self.model.fit,
                      [1, 2, None, 3],
                      x=[1, 2, 3, 4],
                      **self.guess())

    def test_missing_param_raises_error(self):
        # using keyword argument parameters
        guess_missing_sigma = self.guess()
        del guess_missing_sigma['sigma']
        # f = lambda: self.model.fit(self.data, x=self.x, **guess_missing_sigma)
        # self.assertRaises(ValueError, f)

        # using Parameters
        params = self.model.make_params()
        for param_name, value in guess_missing_sigma.items():
            params[param_name].value = value
        self.model.fit(self.data, params, x=self.x)

    def test_extra_param_issues_warning(self):
        # The function accepts extra params, Model will warn but not raise.
        def flexible_func(x, amplitude, center, sigma, **kwargs):
            return gaussian(x, amplitude, center, sigma)

        flexible_model = Model(flexible_func)
        pars = flexible_model.make_params(**self.guess())
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            flexible_model.fit(self.data, pars, x=self.x, extra=5)
        self.assertTrue(len(w) == 1)
        self.assertTrue(issubclass(w[-1].category, UserWarning))

    def test_missing_independent_variable_raises_error(self):
        pars = self.model.make_params(**self.guess())
        f = lambda: self.model.fit(self.data, pars)
        self.assertRaises(ValueError, f)

    def test_independent_var_parsing(self):
        """test the parsing of independent variables for model functions
        with keyword arguments

        step:  form='linear'
        voigt: gamma=None,     can become a variable!!
        """
        stepmod = Model(step)
        assert 'x' in stepmod.independent_vars
        assert 'form' in stepmod.independent_vars
        assert 'linear' == stepmod.independent_vars_defvals.get('form', None)

        voigtmod = Model(voigt)
        assert 'x' in voigtmod.independent_vars
        assert 'gamma' in voigtmod.independent_vars
        assert voigtmod.independent_vars_defvals['gamma'] is None

        pars1 = voigtmod.make_params(amplitude=25, center=9.5, sigma=1)
        assert 'sigma' in pars1
        assert 'gamma' not in pars1

        pars2 = voigtmod.make_params(amplitude=25, center=9.5, sigma=1, gamma=0.5)
        assert 'sigma' in pars2
        assert 'gamma' in pars2

    def test_bounding(self):
        true_values = self.true_values()
        true_values['center'] = 1.3  # as close as it's allowed to get
        pars = self.model.make_params(**self.guess())
        pars['center'].set(value=2, min=1.3)
        result = self.model.fit(self.data, pars, x=self.x)
        assert_results_close(result.values, true_values, rtol=0.05)

    def test_vary_false(self):
        true_values = self.true_values()
        true_values['center'] = 1.3
        pars = self.model.make_params(**self.guess())
        pars['center'].set(value=1.3, vary=False)
        result = self.model.fit(self.data, pars, x=self.x)
        assert_results_close(result.values, true_values, rtol=0.05)

    # testing model addition...

    def test_user_defined_gaussian_plus_constant(self):
        data = self.data + 5.0
        model = self.model + models.ConstantModel()
        guess = self.guess()
        pars = model.make_params(c=10.1, **guess)
        true_values = self.true_values()
        true_values['c'] = 5.0

        result = model.fit(data, pars, x=self.x)
        assert_results_close(result.values, true_values, rtol=0.01, atol=0.01)

    def test_model_with_prefix(self):
        # model with prefix of 'a' and 'b'
        mod = models.GaussianModel(prefix='a')
        vals = {'center': 2.45, 'sigma': 0.8, 'amplitude': 3.15}
        data = gaussian(x=self.x, **vals) + self.noise/3.0
        pars = mod.guess(data, x=self.x)
        self.assertTrue('aamplitude' in pars)
        self.assertTrue('asigma' in pars)
        out = mod.fit(data, pars, x=self.x)
        self.assertTrue(out.params['aamplitude'].value > 2.0)
        self.assertTrue(out.params['acenter'].value > 2.0)
        self.assertTrue(out.params['acenter'].value < 3.0)

        mod = models.GaussianModel(prefix='b')
        data = gaussian(x=self.x, **vals) + self.noise/3.0
        pars = mod.guess(data, x=self.x)
        self.assertTrue('bamplitude' in pars)
        self.assertTrue('bsigma' in pars)

    def test_change_prefix(self):
        "should pass!"
        mod = models.GaussianModel(prefix='b')
        set_prefix_failed = None
        try:
            mod.prefix = 'c'
            set_prefix_failed = False
        except AttributeError:
            set_prefix_failed = True
        except Exception:
            set_prefix_failed = None
        self.assertFalse(set_prefix_failed)

        new_expr = mod.param_hints['fwhm']['expr']
        self.assertTrue('csigma' in new_expr)
        self.assertFalse('bsigma' in new_expr)

    def test_model_name(self):
        # test setting the name for built-in models
        mod = models.GaussianModel(name='user_name')
        self.assertEqual(mod.name, "Model(user_name)")

    def test_sum_of_two_gaussians(self):
        # two user-defined gaussians
        model1 = self.model
        f2 = lambda x, amp, cen, sig: gaussian(x, amplitude=amp, center=cen,
                                               sigma=sig)
        model2 = Model(f2)
        values1 = self.true_values()
        values2 = {'cen': 2.45, 'sig': 0.8, 'amp': 3.15}

        data = (gaussian(x=self.x, **values1) + f2(x=self.x, **values2) +
                self.noise/3.0)
        model = self.model + model2
        pars = model.make_params()
        pars['sigma'].set(value=2, min=0)
        pars['center'].set(value=1, min=0.2, max=1.8)
        pars['amplitude'].set(value=3, min=0)
        pars['sig'].set(value=1, min=0)
        pars['cen'].set(value=2.4, min=2, max=3.5)
        pars['amp'].set(value=1, min=0)

        true_values = dict(list(values1.items()) + list(values2.items()))
        result = model.fit(data, pars, x=self.x)
        assert_results_close(result.values, true_values, rtol=0.01, atol=0.01)

        # user-defined models with common parameter names
        # cannot be added, and should raise
        f = lambda: model1 + model1
        self.assertRaises(NameError, f)

        # two predefined_gaussians, using suffix to differentiate
        model1 = models.GaussianModel(prefix='g1_')
        model2 = models.GaussianModel(prefix='g2_')
        model = model1 + model2
        true_values = {'g1_center': values1['center'],
                       'g1_amplitude': values1['amplitude'],
                       'g1_sigma': values1['sigma'],
                       'g2_center': values2['cen'],
                       'g2_amplitude': values2['amp'],
                       'g2_sigma': values2['sig']}
        pars = model.make_params()
        pars['g1_sigma'].set(2)
        pars['g1_center'].set(1)
        pars['g1_amplitude'].set(3)
        pars['g2_sigma'].set(1)
        pars['g2_center'].set(2.4)
        pars['g2_amplitude'].set(1)

        result = model.fit(data, pars, x=self.x)
        assert_results_close(result.values, true_values, rtol=0.01, atol=0.01)

        # without suffix, the names collide and Model should raise
        model1 = models.GaussianModel()
        model2 = models.GaussianModel()
        f = lambda: model1 + model2
        self.assertRaises(NameError, f)

    def test_sum_composite_models(self):
        # test components of composite model created adding composite model
        model1 = models.GaussianModel(prefix='g1_')
        model2 = models.GaussianModel(prefix='g2_')
        model3 = models.GaussianModel(prefix='g3_')
        model4 = models.GaussianModel(prefix='g4_')

        model_total1 = (model1 + model2) + model3
        for mod in [model1, model2, model3]:
            self.assertTrue(mod in model_total1.components)

        model_total2 = model1 + (model2 + model3)
        for mod in [model1, model2, model3]:
            self.assertTrue(mod in model_total2.components)

        model_total3 = (model1 + model2) + (model3 + model4)
        for mod in [model1, model2, model3, model4]:
            self.assertTrue(mod in model_total3.components)

    def test_eval_components(self):
        model1 = models.GaussianModel(prefix='g1_')
        model2 = models.GaussianModel(prefix='g2_')
        model3 = models.ConstantModel(prefix='bkg_')
        mod = model1 + model2 + model3
        pars = mod.make_params()

        values1 = dict(amplitude=7.10, center=1.1, sigma=2.40)
        values2 = dict(amplitude=12.2, center=2.5, sigma=0.5)
        data = (1.01 + gaussian(x=self.x, **values1) +
                gaussian(x=self.x, **values2) + 0.05*self.noise)

        pars['g1_sigma'].set(2)
        pars['g1_center'].set(1, max=1.5)
        pars['g1_amplitude'].set(3)
        pars['g2_sigma'].set(1)
        pars['g2_center'].set(2.6, min=2.0)
        pars['g2_amplitude'].set(1)
        pars['bkg_c'].set(1.88)

        result = mod.fit(data, params=pars, x=self.x)

        self.assertTrue(abs(result.params['g1_amplitude'].value - 7.1) < 1.5)
        self.assertTrue(abs(result.params['g2_amplitude'].value - 12.2) < 1.5)
        self.assertTrue(abs(result.params['g1_center'].value - 1.1) < 0.2)
        self.assertTrue(abs(result.params['g2_center'].value - 2.5) < 0.2)
        self.assertTrue(abs(result.params['bkg_c'].value - 1.0) < 0.25)

        comps = mod.eval_components(x=self.x)
        assert 'bkg_' in comps

    def test_composite_has_bestvalues(self):
        # test that a composite model has non-empty best_values
        model1 = models.GaussianModel(prefix='g1_')
        model2 = models.GaussianModel(prefix='g2_')

        mod = model1 + model2
        pars = mod.make_params()

        values1 = dict(amplitude=7.10, center=1.1, sigma=2.40)
        values2 = dict(amplitude=12.2, center=2.5, sigma=0.5)
        data = (gaussian(x=self.x, **values1) + gaussian(x=self.x, **values2)
                + 0.1*self.noise)

        pars['g1_sigma'].set(value=2)
        pars['g1_center'].set(value=1, max=1.5)
        pars['g1_amplitude'].set(value=3)
        pars['g2_sigma'].set(value=1)
        pars['g2_center'].set(value=2.6, min=2.0)
        pars['g2_amplitude'].set(value=1)

        result = mod.fit(data, params=pars, x=self.x)

        self.assertTrue(len(result.best_values) == 6)

        self.assertTrue(abs(result.params['g1_amplitude'].value - 7.1) < 0.5)
        self.assertTrue(abs(result.params['g2_amplitude'].value - 12.2) < 0.5)
        self.assertTrue(abs(result.params['g1_center'].value - 1.1) < 0.2)
        self.assertTrue(abs(result.params['g2_center'].value - 2.5) < 0.2)

        for _, par in pars.items():
            assert len(repr(par)) > 5

    @pytest.mark.skipif(not lmfit.model._HAS_MATPLOTLIB,
                        reason="requires matplotlib.pyplot")
    def test_composite_plotting(self):
        # test that a composite model has non-empty best_values
        import matplotlib
        try:
            matplotlib.pyplot.close('all')
        except ValueError:
            pass
        matplotlib.use('Agg')

        model1 = models.GaussianModel(prefix='g1_')
        model2 = models.GaussianModel(prefix='g2_')

        mod = model1 + model2
        pars = mod.make_params()

        values1 = dict(amplitude=7.10, center=1.1, sigma=2.40)
        values2 = dict(amplitude=12.2, center=2.5, sigma=0.5)
        data = (gaussian(x=self.x, **values1) + gaussian(x=self.x, **values2)
                + 0.1*self.noise)

        pars['g1_sigma'].set(2)
        pars['g1_center'].set(1, max=1.5)
        pars['g1_amplitude'].set(3)
        pars['g2_sigma'].set(1)
        pars['g2_center'].set(2.6, min=2.0)
        pars['g2_amplitude'].set(1)

        result = mod.fit(data, params=pars, x=self.x)
        fig = result.plot(show_init=True)

        assert isinstance(fig, matplotlib.figure.Figure)

        comps = result.eval_components(x=self.x)
        assert len(comps) == 2
        assert 'g1_' in comps

    def test_hints_in_composite_models(self):
        # test propagation of hints from base models to composite model
        def func(x, amplitude):
            pass

        m1 = Model(func, prefix='p1_')
        m2 = Model(func, prefix='p2_')

        m1.set_param_hint('amplitude', value=1)
        m2.set_param_hint('amplitude', value=2)

        mx = (m1 + m2)
        params = mx.make_params()
        param_values = {name: p.value for name, p in params.items()}
        self.assertEqual(param_values['p1_amplitude'], 1)
        self.assertEqual(param_values['p2_amplitude'], 2)

    def test_hints_for_peakmodels(self):
        # test that height/fwhm do not cause asteval errors.

        x = np.linspace(-10, 10, 101)
        y = np.sin(x / 3) + x/100.

        m1 = models.LinearModel(prefix='m1_')
        params = m1.guess(y, x=x)

        m2 = models.GaussianModel(prefix='m2_')
        params.update(m2.make_params())

        _m = m1 + m2  # noqa: F841

        param_values = {name: p.value for name, p in params.items()}
        assert_almost_equal(param_values['m1_intercept'], 0.)
        self.assertEqual(param_values['m2_amplitude'], 1)

    def test_weird_param_hints(self):
        # tests Github Issue 312, a very weird way to access param_hints
        def func(x, amp):
            return amp*x

        m = Model(func)
        models = {}
        for i in range(2):
            m.set_param_hint('amp', value=1)
            m.set_param_hint('amp', value=25)

            models[i] = Model(func, prefix=f'mod{i}')
            models[i].param_hints['amp'] = m.param_hints['amp']

        self.assertEqual(models[0].param_hints['amp'],
                         models[1].param_hints['amp'])

    def test_param_hint_explicit_value(self):
        # tests Github Issue 384
        pmod = PseudoVoigtModel()
        params = pmod.make_params(sigma=2, fraction=0.77)
        assert_allclose(params['fraction'].value, 0.77, rtol=0.01)

    def test_symmetric_boundss(self):
        # tests Github Issue 700
        np.random.seed(0)

        x = np.linspace(0, 20, 51)
        y = gaussian(x, amplitude=8.0, center=13, sigma=2.5)
        y += np.random.normal(size=len(x), scale=0.1)

        mod = Model(gaussian)
        params = mod.make_params(sigma=2.2, center=10, amplitude=10)
        # carefully selected to have inexact floating-point representation
        params['sigma'].min = 2.2 - 0.95
        params['sigma'].max = 2.2 + 0.95

        result = mod.fit(y, params, x=x)
        print(result.fit_report())
        self.assertTrue(result.params['sigma'].value > 2.3)
        self.assertTrue(result.params['sigma'].value < 2.7)
        self.assertTrue(result.params['sigma'].stderr is not None)
        self.assertTrue(result.params['amplitude'].stderr is not None)
        self.assertTrue(result.params['sigma'].stderr > 0.02)
        self.assertTrue(result.params['sigma'].stderr < 0.50)

    def test_unprefixed_name_collisions(self):
        # tests Github Issue 710
        np.random.seed(0)
        x = np.linspace(0, 20, 201)
        y = 6 + x * 0.55 + gaussian(x, 4.5, 8.5, 2.1) + np.random.normal(size=len(x), scale=0.03)

        def myline(x, a, b):
            return a + b * x

        def mygauss(x, a, b, c):
            return gaussian(x, a, b, c)

        mod = Model(myline, prefix='line_') + Model(mygauss, prefix='peak_')
        pars = mod.make_params(line_a=5, line_b=1, peak_a=10, peak_b=10, peak_c=5)
        pars.add('a', expr='line_a + peak_a')

        result = mod.fit(y, pars, x=x)
        self.assertTrue(result.params['peak_a'].value > 4)
        self.assertTrue(result.params['peak_a'].value < 5)
        self.assertTrue(result.params['peak_b'].value > 8)
        self.assertTrue(result.params['peak_b'].value < 9)
        self.assertTrue(result.params['peak_c'].value > 1.5)
        self.assertTrue(result.params['peak_c'].value < 2.5)
        self.assertTrue(result.params['line_a'].value > 5.5)
        self.assertTrue(result.params['line_a'].value < 6.5)
        self.assertTrue(result.params['line_b'].value > 0.25)
        self.assertTrue(result.params['line_b'].value < 0.75)
        self.assertTrue(result.params['a'].value > 10)
        self.assertTrue(result.params['a'].value < 11)

    def test_composite_model_with_expr_constrains(self):
        """Smoke test for composite model fitting with expr constraints."""
        y = [0, 0, 4, 2, 1, 8, 21, 21, 23, 35, 50, 54, 46, 70, 77, 87, 98,
             113, 148, 136, 185, 195, 194, 168, 170, 139, 155, 115, 132, 109,
             102, 85, 69, 81, 82, 80, 71, 64, 79, 88, 111, 97, 97, 73, 72, 62,
             41, 30, 13, 3, 9, 7, 0, 0, 0]
        x = np.arange(-0.2, 1.2, 0.025)[:-1] + 0.5*0.025

        def gauss(x, sigma, mu, A):
            return A*np.exp(-(x-mu)**2/(2*sigma**2))

        # Initial values
        p1_mu = 0.2
        p1_sigma = 0.1
        p2_sigma = 0.1

        peak1 = Model(gauss, prefix='p1_')
        peak2 = Model(gauss, prefix='p2_')
        model = peak1 + peak2

        model.set_param_hint('p1_mu', value=p1_mu, min=-1, max=2)
        model.set_param_hint('p1_sigma', value=p1_sigma, min=0.01, max=0.2)
        model.set_param_hint('p2_sigma', value=p2_sigma, min=0.01, max=0.2)
        model.set_param_hint('p1_A', value=100, min=0.01)
        model.set_param_hint('p2_A', value=50, min=0.01)

        # Constrains the distance between peaks to be > 0
        model.set_param_hint('pos_delta', value=0.3, min=0)
        model.set_param_hint('p2_mu', min=-1, expr='p1_mu + pos_delta')

        # Test fitting
        result = model.fit(y, x=x)
        self.assertTrue(result.params['pos_delta'].value > 0)

    def test_model_nan_policy(self):
        """Tests for nan_policy with NaN values in the input data."""
        x = np.linspace(0, 10, 201)
        np.random.seed(0)
        y = gaussian(x, 10.0, 6.15, 0.8)
        y += gaussian(x, 8.0, 6.35, 1.1)
        y += gaussian(x, 0.25, 6.00, 7.5)
        y += np.random.normal(size=len(x), scale=0.5)

        # with NaN values in the input data
        y[55] = y[91] = np.nan
        mod = PseudoVoigtModel()
        params = mod.make_params(amplitude=20, center=5.5,
                                 sigma=1, fraction=0.25)
        params['fraction'].vary = False

        # with raise, should get a ValueError
        result = lambda: mod.fit(y, params, x=x, nan_policy='raise')
        msg = ('NaN values detected in your input data or the output of your '
               'objective/model function - fitting algorithms cannot handle this!')
        self.assertRaisesRegex(ValueError, msg, result)

        # with propagate, should get no error, but bad results
        result = mod.fit(y, params, x=x, nan_policy='propagate')

        # for SciPy v1.10+ this results in an AbortFitException, even with
        # `max_nfev=100000`:
        #   lmfit.minimizer.AbortFitException: fit aborted: too many function
        #   evaluations xxxxx
        if int(scipy_version.split('.')[1]) < 10:
            self.assertTrue(np.isnan(result.chisqr))
            self.assertTrue(np.isnan(result.aic))
            self.assertFalse(result.errorbars)
            self.assertTrue(result.params['amplitude'].stderr is None)
            self.assertTrue(abs(result.params['amplitude'].value - 20.0) < 0.001)
        else:
            pass

        # with omit, should get good results
        result = mod.fit(y, params, x=x, nan_policy='omit')
        self.assertTrue(result.success)
        self.assertTrue(result.chisqr > 2.0)
        self.assertTrue(result.aic < -100)
        self.assertTrue(result.errorbars)
        self.assertTrue(result.params['amplitude'].stderr > 0.1)
        self.assertTrue(abs(result.params['amplitude'].value - 20.0) < 5.0)
        self.assertTrue(abs(result.params['center'].value - 6.0) < 0.5)

        # with 'wrong_argument', should get a ValueError
        err_msg = r"nan_policy must be 'propagate', 'omit', or 'raise'."
        with pytest.raises(ValueError, match=err_msg):
            mod.fit(y, params, x=x, nan_policy='wrong_argument')

    def test_model_nan_policy_NaNs_by_model(self):
        """Test for nan_policy with NaN values generated by the model function."""
        def double_exp(x, a1, t1, a2, t2):
            return a1*np.exp(-x/t1) + a2*np.exp(-(x-0.1) / t2)

        model = Model(double_exp)

        truths = (3.0, 2.0, -5.0, 10.0)
        x = np.linspace(1, 10, 250)
        np.random.seed(0)
        y = double_exp(x, *truths) + 0.1*np.random.randn(x.size)

        p = model.make_params(a1=4, t1=3, a2=4, t2=3)
        result = lambda: model.fit(data=y, params=p, x=x, method='Nelder',
                                   nan_policy='raise')

        msg = 'The model function generated NaN values and the fit aborted!'
        self.assertRaisesRegex(ValueError, msg, result)

    def test_wrapped_model_func(self):
        x = np.linspace(-1, 1, 51)
        y = 2.0*x + 3 + 0.0003 * x*x
        y += np.random.normal(size=len(x), scale=0.025)
        mod = Model(linear_func)
        pars = mod.make_params(a=1.5, b=2.5)

        tmp = mod.eval(pars, x=x)

        self.assertTrue(tmp.max() > 3)
        self.assertTrue(tmp.min() > -20)

        result = mod.fit(y, pars, x=x)
        self.assertTrue(result.chisqr < 0.05)
        self.assertTrue(result.aic < -350)
        self.assertTrue(result.errorbars)

        self.assertTrue(abs(result.params['a'].value - 2.0) < 0.05)
        self.assertTrue(abs(result.params['b'].value - 3.0) < 0.41)

    def test_different_independent_vars_composite_modeld(self):
        """Regression test for different independent variables in CompositeModel.

        See: https://github.com/lmfit/lmfit-py/discussions/787

        """
        def two_independent_vars(y, z, a):
            return a * y + z

        BackgroundModel = Model(two_independent_vars,
                                independent_vars=["y", "z"], prefix="yz_")
        PeakModel = Model(gaussian, independent_vars=["x"], prefix="x_")
        CompModel = BackgroundModel + PeakModel
        assert CompModel.independent_vars == ['x', 'y', 'z']


class TestLinear(CommonTests, unittest.TestCase):

    def setUp(self):
        self.true_values = lambda: dict(slope=5, intercept=2)
        self.guess = lambda: dict(slope=10, intercept=6)
        self.model_constructor = models.LinearModel
        super().setUp()


class TestParabolic(CommonTests, unittest.TestCase):

    def setUp(self):
        self.true_values = lambda: dict(a=5, b=2, c=8)
        self.guess = lambda: dict(a=1, b=6, c=3)
        self.model_constructor = models.ParabolicModel
        super().setUp()


class TestPolynomialOrder2(CommonTests, unittest.TestCase):
    # class Polynomial constructed with order=2
    def setUp(self):
        self.true_values = lambda: dict(c2=5, c1=2, c0=8)
        self.guess = lambda: dict(c1=1, c2=6, c0=3)
        self.model_constructor = models.PolynomialModel
        self.args = (2,)
        super().setUp()


class TestPolynomialOrder3(CommonTests, unittest.TestCase):
    # class Polynomial constructed with order=3
    def setUp(self):
        self.true_values = lambda: dict(c3=2, c2=5, c1=2, c0=8)
        self.guess = lambda: dict(c3=1, c1=1, c2=6, c0=3)
        self.model_constructor = models.PolynomialModel
        self.args = (3,)
        super().setUp()


class TestConstant(CommonTests, unittest.TestCase):
    def setUp(self):
        self.true_values = lambda: dict(c=5)
        self.guess = lambda: dict(c=2)
        self.model_constructor = models.ConstantModel
        super().setUp()

    def check_skip_independent_vars(self):
        raise pytest.skip("ConstantModel has not independent_vars.")


class TestPowerlaw(CommonTests, unittest.TestCase):
    def setUp(self):
        self.true_values = lambda: dict(amplitude=5, exponent=3)
        self.guess = lambda: dict(amplitude=2, exponent=8)
        self.model_constructor = models.PowerLawModel
        super().setUp()


class TestExponential(CommonTests, unittest.TestCase):
    def setUp(self):
        self.true_values = lambda: dict(amplitude=5, decay=3)
        self.guess = lambda: dict(amplitude=2, decay=8)
        self.model_constructor = models.ExponentialModel
        super().setUp()


class TestComplexConstant(CommonTests, unittest.TestCase):
    def setUp(self):
        self.true_values = lambda: dict(re=5, im=5)
        self.guess = lambda: dict(re=2, im=2)
        self.model_constructor = models.ComplexConstantModel
        super().setUp()


class TestExpression(CommonTests, unittest.TestCase):
    def setUp(self):
        self.true_values = lambda: dict(off_c=0.25, amp_c=1.0, x0=2.0)
        self.guess = lambda: dict(off_c=0.20, amp_c=1.5, x0=2.5)
        self.expression = "off_c + amp_c * exp(-x/x0)"
        self.model_constructor = (
            lambda *args, **kwargs: models.ExpressionModel(self.expression, *args, **kwargs))
        super().setUp()

    def test_composite_with_expression(self):
        expression_model = models.ExpressionModel("exp(-x/x0)", name='exp')
        amp_model = models.ConstantModel(prefix='amp_')
        off_model = models.ConstantModel(prefix='off_', name="off")

        comp_model = off_model + amp_model * expression_model

        x = self.x
        true_values = self.true_values()
        data = comp_model.eval(x=x, **true_values) + self.noise
        # data = 0.25 + 1 * np.exp(-x / 2.)

        params = comp_model.make_params(**self.guess())

        result = comp_model.fit(data, x=x, params=params)
        assert_results_close(result.values, true_values, rtol=0.01, atol=0.01)

        data_components = comp_model.eval_components(x=x)
        self.assertIn('exp', data_components)


def test_make_params_valuetypes():
    mod = lmfit.models.SineModel()

    pars = mod.make_params(amplitude=1, frequency=1, shift=-0.2)

    pars = mod.make_params(amplitude={'value': 0.9, 'min': 0},
                           frequency=1.03,
                           shift={'value': -0.2, 'vary': False})

    val_i32 = np.arange(10, dtype=np.int32)
    val_i64 = np.arange(10, dtype=np.int64)
    # np.longdouble equals to np.float128 on Linux and macOS, np.float64 on Windows
    val_ld = np.arange(10, dtype=np.longdouble)/3.0
    val_c128 = np.arange(10, dtype=np.complex128)/3.0

    pars = mod.make_params(amplitude=val_i64[2],
                           frequency=val_i32[3],
                           shift=-val_ld[4])

    pars = mod.make_params(amplitude=val_c128[2],
                           frequency=val_i32[3],
                           shift=-val_ld[4])

    assert pars is not None
    with pytest.raises(ValueError):
        pars = mod.make_params(amplitude='a string', frequency=2, shift=7)

    with pytest.raises(TypeError):
        pars = mod.make_params(amplitude={'v': 3}, frequency=2, shift=7)

    with pytest.raises(TypeError):
        pars = mod.make_params(amplitude={}, frequency=2, shift=7)


def test_complex_model_eval_uncertainty():
    """Github #900"""
    def cmplx(f, omega, areal, aimag, off, sigma):
        return (areal*np.cos(f*omega + off) + 1j*aimag*np.sin(f*omega + off))*np.exp(-f/sigma)

    f = np.linspace(0, 10, 501)
    dat = cmplx(f, 4, 10, 5, 0.2, 4.5) + (0.1 + 0.2j)*np.random.normal(scale=0.25, size=len(f))
    mod = Model(cmplx)
    params = mod.make_params(omega=5, areal=5, aimag=5,
                             off={'value': 0.5, 'min': -2, 'max': 2},
                             sigma={'value': 3, 'min': 1.e-5, 'max': 1000})

    result = mod.fit(dat, params=params, f=f)
    dfit = result.eval_uncertainty()
    assert len(dfit) == len(f)
    assert dfit.dtype == 'complex128'


def test_compositemodel_returning_list():
    """Github #875"""
    def lin1(x, k):
        return [k*x1 for x1 in x]

    def lin2(x, k):
        return [k*x1 for x1 in x]

    y = np.linspace(0, 100, 100)
    x = np.linspace(0, 100, 100)

    Model1 = Model(lin1, independent_vars=["x"], prefix="m1_")
    Model2 = Model(lin2, independent_vars=["x"], prefix="m2_")
    ModelSum = Model1 + Model2
    pars = Parameters()
    pars.add('m1_k', value=0.5)
    pars.add('m2_k', value=0.5)
    result = ModelSum.fit(y, pars, x=x)
    assert len(result.best_fit) == len(x)


def test_rsquared_with_weights():
    """Github #921"""
    def func(x, k=1, b=0):
        return k*x+b

    x = np.array([1, 2, 3, 4])
    y = np.array([1.1, 1.9, 3.05, 3.95])
    yerr = np.array([0.03, 0.04, 0.01, 0.02])

    mod = Model(func)
    params = mod.make_params()
    result = mod.fit(y, params, x=x, weights=1.0/yerr)

    assert result.rsquared < 1.00
    assert result.rsquared > 0.95



"""
Fit Two Dimensional Peaks
=========================

This example illustrates how to handle two-dimensional data with lmfit.

"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import griddata

import lmfit
from lmfit.lineshapes import gaussian2d, lorentzian

###############################################################################
# Two-dimensional Gaussian
# ------------------------
# We start by considering a simple two-dimensional gaussian function, which
# depends on coordinates `(x, y)`. The most general case of experimental
# data will be irregularly sampled and noisy. Let's simulate some:
npoints = 10000
np.random.seed(2021)
x = np.random.rand(npoints)*10 - 4
y = np.random.rand(npoints)*5 - 3
z = gaussian2d(x, y, amplitude=30, centerx=2, centery=-.5, sigmax=.6, sigmay=.8)
z += 2*(np.random.rand(*z.shape)-.5)
error = np.sqrt(z+1)
###############################################################################
# To plot this, we can interpolate the data onto a grid.
X, Y = np.meshgrid(np.linspace(x.min(), x.max(), 100),
                   np.linspace(y.min(), y.max(), 100))
Z = griddata((x, y), z, (X, Y), method='linear', fill_value=0)

fig, ax = plt.subplots()
art = ax.pcolor(X, Y, Z, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_xlabel('x')
ax.set_ylabel('y')
plt.show()
###############################################################################
# In this case, we can use a built-in model to fit
model = lmfit.models.Gaussian2dModel()
params = model.guess(z, x, y)
result = model.fit(z, x=x, y=y, params=params, weights=1/error)
lmfit.report_fit(result)
###############################################################################
# To check the fit, we can evaluate the function on the same grid we used
# before and make plots of the data, the fit and the difference between the two.
fig, axs = plt.subplots(2, 2, figsize=(10, 10))

vmax = np.nanpercentile(Z, 99.9)

ax = axs[0, 0]
art = ax.pcolor(X, Y, Z, vmin=0, vmax=vmax, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Data')

ax = axs[0, 1]
fit = model.func(X, Y, **result.best_values)
art = ax.pcolor(X, Y, fit, vmin=0, vmax=vmax, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Fit')

ax = axs[1, 0]
fit = model.func(X, Y, **result.best_values)
art = ax.pcolor(X, Y, Z-fit, vmin=0, vmax=10, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Data - Fit')

for ax in axs.ravel():
    ax.set_xlabel('x')
    ax.set_ylabel('y')
axs[1, 1].remove()
plt.show()


###############################################################################
# Two-dimensional off-axis Lorentzian
# -----------------------------------
# We now go on to show a harder example, in which the peak has a Lorentzian
# profile and an off-axis anisotropic shape. This can be handled by applying a
# suitable coordinate transform and then using the `lorentzian` function that
# lmfit provides in the lineshapes module.
def lorentzian2d(x, y, amplitude=1., centerx=0., centery=0., sigmax=1., sigmay=1.,
                 rotation=0):
    """Return a two dimensional lorentzian.

    The maximum of the peak occurs at ``centerx`` and ``centery``
    with widths ``sigmax`` and ``sigmay`` in the x and y directions
    respectively. The peak can be rotated by choosing the value of ``rotation``
    in radians.
    """
    xp = (x - centerx)*np.cos(rotation) - (y - centery)*np.sin(rotation)
    yp = (x - centerx)*np.sin(rotation) + (y - centery)*np.cos(rotation)
    R = (xp/sigmax)**2 + (yp/sigmay)**2

    return 2*amplitude*lorentzian(R)/(np.pi*sigmax*sigmay)


###############################################################################
# Data can be simulated and plotted in the same way as we did before.
npoints = 10000
x = np.random.rand(npoints)*10 - 4
y = np.random.rand(npoints)*5 - 3
z = lorentzian2d(x, y, amplitude=30, centerx=2, centery=-.5, sigmax=.6,
                 sigmay=1.2, rotation=30*np.pi/180)
z += 2*(np.random.rand(*z.shape)-.5)
error = np.sqrt(z+1)

X, Y = np.meshgrid(np.linspace(x.min(), x.max(), 100),
                   np.linspace(y.min(), y.max(), 100))
Z = griddata((x, y), z, (X, Y), method='linear', fill_value=0)

fig, ax = plt.subplots()
ax.set_xlabel('x')
ax.set_ylabel('y')
art = ax.pcolor(X, Y, Z, shading='auto')
plt.colorbar(art, ax=ax, label='z')
plt.show()

###############################################################################
# To fit, create a model from the function. Don't forget to tell lmfit that both
# `x` and `y` are independent variables. Keep in mind that lmfit will take the
# function keywords as default initial guesses in this case and that it will not
# know that certain parameters only make physical sense over restricted ranges.
# For example, peak widths should be positive and the rotation can be restricted
# over a quarter circle.
model = lmfit.Model(lorentzian2d, independent_vars=['x', 'y'])
params = model.make_params(amplitude=10, centerx=x[np.argmax(z)],
                           centery=y[np.argmax(z)])
params['rotation'].set(value=.1, min=0, max=np.pi/2)
params['sigmax'].set(value=1, min=0)
params['sigmay'].set(value=2, min=0)

result = model.fit(z, x=x, y=y, params=params, weights=1/error)
lmfit.report_fit(result)

###############################################################################
# The process of making plots to check it worked is the same as before.
fig, axs = plt.subplots(2, 2, figsize=(10, 10))

vmax = np.nanpercentile(Z, 99.9)

ax = axs[0, 0]
art = ax.pcolor(X, Y, Z, vmin=0, vmax=vmax, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Data')

ax = axs[0, 1]
fit = model.func(X, Y, **result.best_values)
art = ax.pcolor(X, Y, fit, vmin=0, vmax=vmax, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Fit')

ax = axs[1, 0]
fit = model.func(X, Y, **result.best_values)
art = ax.pcolor(X, Y, Z-fit, vmin=0, vmax=10, shading='auto')
plt.colorbar(art, ax=ax, label='z')
ax.set_title('Data - Fit')

for ax in axs.ravel():
    ax.set_xlabel('x')
    ax.set_ylabel('y')
axs[1, 1].remove()
plt.show()



"""Test speed of building complex model."""
from copy import deepcopy
import sys
import time

import numpy as np
import pytest

from lmfit import Model
from lmfit.lineshapes import gaussian

sys.setrecursionlimit(2000)


@pytest.mark.flaky(max_runs=5)
def test_manypeaks_speed():
    model = None
    t0 = time.time()
    for i in np.arange(500):
        g = Model(gaussian, prefix=f'g{i}')
        if model is None:
            model = g
        else:
            model += g
    t1 = time.time()
    pars = model.make_params()
    t2 = time.time()
    _cpars = deepcopy(pars)  # noqa: F841
    t3 = time.time()

    # these are very conservative tests that
    # should be satisfied on nearly any machine
    assert (t3-t2) < 0.5
    assert (t2-t1) < 0.5
    assert (t1-t0) < 5.0
